# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/instance_config_instance_switcher.ipynb.

# %% auto 0
__all__ = ['InstanceSwitcherMapping_GET_Error', 'get_instance_switcher_mapping', 'InstanceSwitcherMapping_CRUD_Error',
           'set_instance_switcher_mapping']

# %% ../../nbs/routes/instance_config_instance_switcher.ipynb 2
import httpx

from typing import List

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de


# %% ../../nbs/routes/instance_config_instance_switcher.ipynb 6
class InstanceSwitcherMapping_GET_Error(de.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        message=None,
    ):

        super().__init__(
            res=res,
            message=message or res.response,
        )


# gets existing instance switcher mapping, response = list[dict]
@gd.route_function
async def get_instance_switcher_mapping(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop: int = 1,
    timeout: int = 20
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/everywhere/admin/userattributeinstances"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        timeout=timeout
    )

    if not res.is_success:
        raise InstanceSwitcherMapping_GET_Error(
            res=res,
            message= f"failed to retrieve instance switcher mapping - {res.response}",
        )

    return res

# %% ../../nbs/routes/instance_config_instance_switcher.ipynb 9
class InstanceSwitcherMapping_CRUD_Error(de.RouteError):
    def __init__(
        self,
        res,
        message=None,
    ):
        super().__init__(
            res=res,
            message=message or res.response,
        )


# update the instance switcher mappings
@gd.route_function
async def set_instance_switcher_mapping(
    auth: dmda.DomoAuth,
    mapping_payloads: List[dict],
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop: int = 1,
    timeout: int = 60
) -> rgd.ResponseGetData:
    """
    accepts a list of instance_switcher_mappings format:
    mapping_payloads = [ {'userAttribute': 'test1','instance': 'test.domo.com'}]

    """

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/everywhere/admin/userattributeinstances"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        body=mapping_payloads,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        timeout=timeout
    )

    if not res.is_success:
        raise InstanceSwitcherMapping_CRUD_Error(
            res=res,
            message=f"failed to update instance switcher mappings - {res.response}",
        )

    res.response = f"success: updated instance switcher mappings"
    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/CodeEngine.ipynb.

# %% auto 0
__all__ = ['ExportExtension', 'DomoCodeEngine_ConfigError', 'DomoCodeEngine_PackageVersion', 'DomoCodeEngine_Package']

# %% ../../nbs/classes/CodeEngine.ipynb 2
# from domolibrary.routes.codeengine import extract_functions
from .CodeEngineManifest import CodeEngineManifest, CodeEngineManifest_Function, CodeEngineManifest_Argument, CodeEngine_Argument, PythonTypeToSchemaType


# %% ../../nbs/classes/CodeEngine.ipynb 3
import os

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Callable
import datetime as dt

from io import open
import httpx

from nbdev.showdoc import patch_to

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoEntity as dmee    

import domolibrary.routes.codeengine as codeengine_routes
import domolibrary.client.DomoError as dmde
import domolibrary.utils.files as dmuf
from domolibrary.utils.convert import convert_string_to_datetime
import domolibrary.classes.DomoUser as dmdu




# %% ../../nbs/classes/CodeEngine.ipynb 7
class ExportExtension(Enum):
    JAVASCRIPT = "js"
    PYTHON = "py"
    
class DomoCodeEngine_ConfigError(dmde.ClassError):
    def __init__(
        self,
        cls_instance,
        package_id,
        version=None,
        message=None,
        ):
        super().__init__(
            cls_instance=cls_instance,
            entity_id=package_id,
            message=f"verion {version} | {message}" if version else message,
        )
    

# %% ../../nbs/classes/CodeEngine.ipynb 9
@dataclass
class DomoCodeEngine_PackageVersion:
    auth: dmda.DomoAuth = field(repr=False)
    package_id: str
    version: str
    language: str
    description: str
    createdby_id: int
    released_dt: dt.datetime
    configuration: dict

    createdby: dmdu.DomoUser = None
    accounts_mapping: List[int] = None
    ml_model: List[str] = None

    code: str = field(repr=False, default=None)
    # functions_ls: List[CodeEngineManifest_Function] = field(repr=False, default=None)
    functions: dict = None

    Manifest : CodeEngineManifest = field(default=None)

    def _set_configuration(self, configuration=None):
        if configuration:
            self.configuration = configuration

        if not self.configuration:
            raise DomoCodeEngine_ConfigError(
                cls_instance=self,
                package_id=self.package_id,
                version=self.version,
                message="unable to set configuration",
            )

        self.accounts_mapping = self.configuration["accountsMapping"]
        self.ml_model = self.configuration["mlModel"]

        return self

    # def set_functions(
    #     self,
    #     code: str = None,
    #     functions: List[dict] = None,
    #     language: str = None,
    #     # function_parser_fn: Callable = None,  # must receive code string, function_ls, language
    # ):
    #     self.code = code or self.code

    #     self.functions_ls = functions or self.functions_ls

    #     self.language = language or self.language

    #     if not self.code or not self.configuration or not self.language:
    #         raise DomoCodeEngine_ConfigError(
    #             cls_instance=self,
    #             package_id=self.package_id,
    #             version=self.version,
    #             message="API did not return code, configuration or language -- unable to set functions",
    #         )
        # try:
            # function_parser_fn = function_parser_fn or extract_functions
            
            # functions_ls = function_parser_fn(
            #     code=self.code, 
            #     function_ls=self.functions_ls, 
            #     language=self.language
            # )

            # self.functions = [
            #     DomoCodeEngine_Function._from_json(
            #         obj,
            #         is_private= obj.pop("is_private", False),
            #         language=self.language,
            #         package_id=self.package_id,
            #         version=self.version,
            #         code_raw=self.code,
            #     )
            #     for obj in functions_ls
            # ]
            # return functions
        
        # except Exception as e:
        #     print(e)
        #     raise DomoCodeEngine_ConfigError(
        #         package_id=self.package_id,
        #         version=self.version,
        #         message=e,
        #         domo_instance=self.auth.domo_instance,
        #     )

    @classmethod
    def _from_json(
        cls,
        obj: dict,
        package_id,
        auth: dmda.DomoAuth,
        language: str,
        is_supress_error: bool = True,
    ):

        language = (language or obj.get("language", None)).upper()

        domo_version = cls(
            auth=auth,
            package_id=package_id,
            language=language,
            version=obj["version"],
            code=obj.get("code"),
            description=obj.get("description", None),
            createdby_id=obj["createdBy"],
            released_dt=convert_string_to_datetime(obj.get("releasedOn")),
            configuration=obj["configuration"],
            # functions_ls=obj["functions"],
            Manifest = CodeEngineManifest.from_api(obj = obj) if language == "PYTHON" else None 
        )

        # try:
        domo_version._set_configuration()

        # except DomoCodeEngine_ConfigError as e:
        #     print(e)
        #     if not is_supress_error:
        #         raise e from e

        # try:
        #     domo_version.set_functions()
        # except DomoCodeEngine_ConfigError as e:
        #     print(e)
        #     if not is_supress_error:
        #         raise e from e

        return domo_version

    @classmethod
    async def get_by_id_and_version(
        cls,
        package_id,
        version,
        auth: dmda.DomoAuth,
        language: str = None,
        params: dict = None , # {"parts": "functions,code"}
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
    ):
        params = params or {"parts": "functions,code"}

        res = await codeengine_routes.get_codeengine_package_by_id_and_version(
            package_id=package_id,
            version=version,
            auth=auth,
            params = params,
            debug_api=debug_api,
            session=session,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return cls._from_json(
            auth=auth, obj=res.response, 
            package_id=package_id, 
            language=language
        )

    def __eq__(self, other):
        if not isinstance(other, DomoCodeEngine_PackageVersion):
            return False

        if self.version != other.version or self.package_id != other.package_id:
            return False

        return True

    async def download_source_code(
        self,
        download_folder: str = "./EXPORT/codeengine",
        file_name: str = None,
        debug_api: bool = False,
        replace_folder : bool = False
        
    ):
        if not self.code:
            self.code = (await self.get_by_id_and_version(
                auth=self.auth,
                package_id=self.package_id,
                version=self.version,
                debug_api=debug_api,
                params = {"parts" : "code"} 
            )).code

        extension = '.txt'
        if self.language == "PYTHON":
            extension = '.py'

        if self.language == "JAVASCRIPT":
            extension = '.js'
            
        file_name = file_name or f'{self.package_id}/{self.version}/functions{extension}'

        file_path = os.path.join(download_folder, file_name)

        if self.Manifest:
            return self.Manifest.download_source_code(
                export_folder=os.path.join(download_folder, file_path), replace_folder=replace_folder)

        dmuf.upsert_file(file_path, content = self.code, replace_folder=replace_folder)



# %% ../../nbs/classes/CodeEngine.ipynb 13
@patch_to(DomoCodeEngine_PackageVersion)
def export(
    self,
    file_name: str = None,
    output_folder: str = "EXPORT/code_engine/",
    debug_prn: bool = False,
):
    output_folder = (
        f"{output_folder}/" if not output_folder.endswith("/") else output_folder
    )

    dmuf.upsert_folder(output_folder)

    file_name = file_name or self.package_id
    file_name = dmuf.change_extension(file_name, ExportExtension[self.language].value)

    file_path = os.path.join(output_folder, file_name)

    if debug_prn:
        print(output_folder, file_name)

    with open(file_path, "w+", newline="\n", encoding="utf-8") as f:
        f.write(self.code)

    return file_path

# %% ../../nbs/classes/CodeEngine.ipynb 15
@dataclass
class DomoCodeEngine_Package:
    auth: dmda.DomoAuth = field(repr=False)
    id: str
    name: str
    description: str
    language: str
    environment: str
    availability: str
    owner_id: int
    created: dt.datetime
    last_modified: dt.datetime
    functions: List

    current_version: str = None
    versions: List[DomoCodeEngine_PackageVersion] = None
    owner: List[dmdu.DomoUser] = None

    def __post_init__(self):
        self._set_current_version()

    def _set_current_version(self):
        if not self.versions:
            return

        versions = [version.version for version in self.versions]
        self.current_version = max(versions)

    @classmethod
    def from_packages_api(cls, obj, auth):
        package_id = obj["id"]
        language = obj["language"]
        versions = [
            DomoCodeEngine_PackageVersion._from_json(
                version, package_id=package_id, language=language, auth=auth
            )
            for version in obj.get("versions")
        ]

        return cls(
            auth=auth,
            id=package_id,
            name=obj["name"],
            description=obj.get("description"),
            language=language,
            environment=obj.get("environment"),
            availability=obj.get("availability"),
            owner_id=obj.get("owner"),
            versions=versions,
            created=convert_string_to_datetime(obj.get("createdOn")),
            last_modified=convert_string_to_datetime(obj.get("updatedOn")),
            functions=obj.get("functions"),
        )

    @classmethod
    async def get_by_id(
        cls,
        auth: dmda.DomoAuth,
        package_id,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        res = await codeengine_routes.get_codeengine_package_by_id(
            package_id,
            auth=auth,
            debug_api=debug_api,
            params=None,
            session=session,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
        if return_raw:
            return res

        return cls.from_packages_api(res.response, auth=auth)

    def __eq__(self, other):
        if not isinstance(other, DomoCodeEngine_Package):
            return False

        if self.id != other.id:
            return False

        return True

# %% ../../nbs/classes/CodeEngine.ipynb 17
@patch_to(DomoCodeEngine_Package, cls_method=True)
async def get_current_version_by_id(
    cls,
    auth: dmda.DomoAuth,
    package_id,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
):
    domo_package = await cls.get_by_id(
        package_id=package_id,
        auth=auth,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    current_version = domo_package.current_version

    if not domo_package.current_version:
        raise DomoCodeEngine_ConfigError(
            package_id=package_id,
            version=None,
            message="No current version found for the package",
            domo_instance=auth.domo_instance,
        )

    domo_version = await DomoCodeEngine_PackageVersion.get_by_id_and_version(
        package_id=package_id,
        version=current_version,
        auth=auth,
        language=domo_package.language,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    return domo_version



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/CodeEngineManifest.ipynb.

# %% auto 0
__all__ = []

# %% ../../nbs/classes/CodeEngineManifest.ipynb 1
from .CodeEngineManifest_Argument import CodeEngine_Argument, PythonTypeToSchemaType

# %% ../../nbs/classes/CodeEngineManifest.ipynb 2
import domolibrary.utils.convert as dmcv

from typing import Dict, Any
from dataclasses import dataclass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/appdb.ipynb.

# %% auto 0
__all__ = ['AppDb_GET_Exception', 'AppDb_CRUD_Exception', 'get_datastores', 'get_datastore_by_id',
           'get_collections_from_datastore', 'create_datastore', 'create_collection', 'get_collections',
           'get_collection_by_id', 'get_documents_from_collection', 'Collection_Permission_Enum',
           'modify_collection_permissions', 'get_collection_document_by_id', 'create_document', 'update_document']

# %% ../../nbs/routes/appdb.ipynb 2
import httpx
import domolibrary.client.get_data as gd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de

from domolibrary.client.DomoEntity import DomoEnum

# %% ../../nbs/routes/appdb.ipynb 4
class AppDb_GET_Exception(de.DomoError):
    def __init__(self, res, message=None):
        super().__init__(
            status=res.status,
            message=message or res.response,
            domo_instance=res.auth.domo_instance,
            function_name=res.traceback_details.function_name,
            parent_class=res.traceback_details.parent_class,
        )


class AppDb_CRUD_Exception(de.DomoError):
    def __init__(self, res, message=None):
        super().__init__(
            status=res.status,
            message=message or res.response,
            domo_instance=res.auth.domo_instance,
            function_name=res.traceback_details.function_name,
            parent_class=res.traceback_details.parent_class,
        )

# %% ../../nbs/routes/appdb.ipynb 7
gd.route_function


async def get_datastores(
    auth: dmda.DomoAuth,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 9
gd.route_function


async def get_datastore_by_id(
    auth: dmda.DomoAuth,
    datastore_id: str,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 11
gd.route_function


async def get_collections_from_datastore(
    auth: dmda.DomoAuth,
    datastore_id: str,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}/collections"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 14
@gd.route_function
async def create_datastore(
    auth: dmda.DomoAuth,
    datastore_name=None,  # in UI shows up as appName,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/"

    body = {"name": datastore_name}

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_CRUD_Exception(res)

    return res


@gd.route_function
async def create_collection(
    auth: dmda.DomoAuth,
    datastore_id: str,  # collections must be created inside a datastore which will show as the associated app_name
    collection_name: str,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}/collections"

    body = {"name": collection_name}

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_CRUD_Exception(res)

    return res


# NOT SUPPORTED
# @gd.route_function
# async def update_datastore_name(
#     auth: dmda.DomoAuth,
#     datastore_id :str,
#     datastore_name = None, # in UI shows up as appName,
#     parent_class: str = None,
#     debug_api: bool = False,
#     session: httpx.AsyncClient = None,
#     debug_num_stacks_to_drop=1,

# ):

#     url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/{datastore_id}"

#     body = {"name": datastore_name}


#     res = await gd.get_data(
#         auth=auth,
#         method="PUT",
#         url=url,
#         body = body,
#         parent_class=parent_class,
#         debug_api=debug_api,
#         session=session,
#         num_stacks_to_drop=debug_num_stacks_to_drop,
#     )

#     if not res.is_success:
#         raise AppDb_CRUD_Exception(res)

#     return res

# %% ../../nbs/routes/appdb.ipynb 16
gd.route_function


async def get_collections(
    auth: dmda.DomoAuth,
    datastore_id: str = None,  # filters for a specific datastoreId
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/collections/"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        params={"datastoreId": datastore_id},
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success and res.status == 400:
        raise AppDb_GET_Exception(
            res,
            message=f"invalid datastoreId  - {datastore_id} or ensure it's shared with authenticated user",
        )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 18
@gd.route_function
async def get_collection_by_id(
    auth: dmda.DomoAuth,
    collection_id,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/collections/{collection_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 20
@gd.route_function
async def get_documents_from_collection(
    auth: dmda.DomoAuth,
    collection_id: str,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    body=None,
    debug_num_stacks_to_drop=1,
    query : dict = None
):
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v2/collections/{collection_id}/documents/query"

    query = query or {}

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=query,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(
            res, message=f"unable to query documents in collection - {collection_id}"
        )

    return res

# %% ../../nbs/routes/appdb.ipynb 23
class Collection_Permission_Enum(DomoEnum):
    READ_CONTENT = "READ_CONTENT"
    ADMIN = "ADMIN"
    UPDATE_CONTENT = "UPDATE_CONTENT"


@gd.route_function
async def modify_collection_permissions(
    collection_id: str,
    auth: dmda.DomoAuth,
    user_id: str = None,
    group_id: str = None,
    permission=Collection_Permission_Enum.READ_CONTENT,
    debug_api: bool = False,
    return_raw: bool = False,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/collections/{collection_id}/permission/{'USER' if user_id else {'GROUP'}}/{user_id or group_id}"

    params = {
        "overwrite": False,
        "permissions": (
            permission.value
            if isinstance(permission, Collection_Permission_Enum)
            else permission
        ),
    }

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        params=params,
        url=url,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise AppDb_CRUD_Exception(
            res,
            message="unable to set permissions for {user_id} to {permission.value} in collection {collection_id}",
        )

    res.response = f"set permissions for {user_id} to {permission.value} in collection {collection_id}"

    return res

# %% ../../nbs/routes/appdb.ipynb 27
@gd.route_function
async def get_collection_document_by_id(
    auth: dmda.DomoAuth,
    collection_id: str,
    document_id: str,
    query: dict = None,
    parent_class: str = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=1,
):

    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/collections/{collection_id}/documents/{document_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        body=query,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_GET_Exception(res)

    return res

# %% ../../nbs/routes/appdb.ipynb 30
@gd.route_function
async def create_document(
    auth: dmda.DomoAuth,
    collection_id: str,
    content: dict,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop: int = 1,
):
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v1/collections/{collection_id}/documents"

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body={"content": content},
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_CRUD_Exception(res)

    return res


@gd.route_function
async def update_document(
    auth: dmda.DomoAuth,
    collection_id: str,
    document_id: str,
    content: dict,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop: int = 1,
):
    url = f"https://{auth.domo_instance}.domo.com/api/datastores/v2/collections/{collection_id}/documents/{document_id}"

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body={"content": content},
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise AppDb_CRUD_Exception(res)

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoPublish.ipynb.

# %% ../../nbs/classes/50_DomoPublish.ipynb 2
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List, Any
from domolibrary.client.DomoEntity import DomoEnum, DomoEntity_w_Lineage

import datetime as dt
import pandas as pd
import asyncio
import httpx

import uuid

from nbdev.showdoc import patch_to

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoEntity as dmen

import domolibrary.routes.publish as publish_routes

import domolibrary.classes.DomoLineage as dmdl

import domolibrary.utils.chunk_execution as dmce
import domolibrary.client.DomoError as dmde


# %% auto 0
__all__ = ['DomoPublication_Subscription', 'DomoPublication_Content_Enum', 'DomoPublication_Content',
           'DomoPublication_UnexpectedContentType', 'DomoPublication', 'DomoPublications']

# %% ../../nbs/classes/50_DomoPublish.ipynb 6
@dataclass
class DomoPublication_Subscription(dmen.DomoSubEntity):
    """the subscriber represents a location a publication is sent to"""


    subscription_id: str
    publication_id: str
    domain: str
    created_dt: Optional[dt.datetime] = (None,)


    @classmethod
    def _from_json(cls, obj, auth: dmda.DomoAuth, parent : Any= None):

        return cls(
            auth=auth,
            subscription_id=obj.get("id") or obj.get("subscriptionId"),
            publication_id=obj["publicationId"],
            domain=obj.get("domain") or obj.get("publisherDomain"),
            created_dt=(
                (dt.datetime.fromtimestamp(obj.get("created") / 1000))
                if obj.get("created")
                else None
            ),
            parent = parent,
            parent_id = (parent and parent.id) or None
        )

    async def get_content_details(
        self,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):

        if not self.parent:
            raise dmde.DomoError(
                """The .parent attribute (DomoPublication) is required to get content details.
                must call from the publication instance.  Cannot call from the subscriber"""
            )
        
        publication_content = self.parent.content

        res = await publish_routes.get_subscriber_content_details(
            auth=self.auth,
            publication_id=self.publication_id,
            subscriber_instance=self.domain,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            debug_api=debug_api,
            session = session,
            parent_class=self.__class__.__name__,
        )

        for content in publication_content:
            subscriber_obj = next(
                (
                    subscriber_obj
                    for subscriber_obj in res.response
                    if subscriber_obj["publisherObjectId"] == content.entity_id
                    and subscriber_obj["contentType"] == content.entity_type
                ),
                None,
            )
            if subscriber_obj is not None:
                content.subscriber_content_id = subscriber_obj["subscriberObjectId"]
                content.subscriber_insance = subscriber_obj["subscriberDomain"]

        self.content = publication_content

        return self.content

# %% ../../nbs/classes/50_DomoPublish.ipynb 9
class DomoPublication_Content_Enum(DomoEnum):
    import domolibrary.classes.DomoDataset as dmdc
    import domolibrary.classes.DomoCard as dmac
    import domolibrary.classes.DomoPage as dmpg
    import domolibrary.classes.DomoAppStudio as dmas

    CARD = dmac.DomoCard
    DATASET = dmdc.DomoDataset
    DATA_APP = dmas.DomoAppStudio
    PAGE = dmpg.DomoPage



# %% ../../nbs/classes/50_DomoPublish.ipynb 10
@dataclass
class DomoPublication_Content:
    auth: dmda.DomoAuth

    content_id: str
    entity_type: str
    entity_id: str
    entity_domain: str
    is_v2: bool
    is_direct_content: bool

    created_dt: dt.datetime
    updated_dt: dt.datetime = None

    subscriber_content_id: str = None
    subscriber_insance: str = None

    entity: Any = field(repr=False, default=None)
    parent: Any = field(repr=False, default=None)

    """the publication content is the content from the publisher instance that is being distributed to subscribers"""

    @classmethod
    def _from_json(cls, obj: dict, auth: dmda.DomoAuth, parent: Any = None):

        entity_type = obj.get("content").get("type")
        return cls(
            auth=auth,
            content_id=obj["id"],
            entity_id=obj.get("content").get("domoObjectId"),
            entity_domain=obj.get("content").get("domain"),
            is_v2=obj.get("isV2"),
            created_dt=(
                dt.datetime.fromtimestamp(obj["created"] / 1000)
                if obj["created"]
                else None
            ),
            updated_dt=(
                dt.datetime.fromtimestamp(obj.get("content").get("updated") / 1000)
                if obj.get("content").get("updated")
                else None
            ),
            is_direct_content=obj.get("useDirectContent"),
            parent=parent,
            entity_type=entity_type,
            entity=DomoPublication_Content_Enum[entity_type].value,
        )

    async def get_entity(self, debug_api: bool = False, session: httpx.AsyncClient = None):
        """get the entity from the publication content"""
        if not self.entity:
            self.entity=DomoPublication_Content_Enum[self.entity_type].value

        self.entity = await self.entity._get_entity_by_id(
            auth=self.auth,
            entity_id=self.entity_id,
            debug_api=debug_api,
            session = session
        )

        return self.entity

    def to_api_json(self):
        return {
            "domain": self.entity_domain,
            "domoObjectId": self.entity_id,
            "customerId": self.entity_domain,
            "type": self.entity_type,
        }

# %% ../../nbs/classes/50_DomoPublish.ipynb 12
class DomoPublication_UnexpectedContentType(dmde.ClassError):
    def __init__(self, cls_instance, content_type):
        super().__init__(
            cls_instance=cls_instance,
            message=f"DomoPublication_Instantiation: Unexpected content type {content_type}",
        )


@dataclass
class DomoPublication(DomoEntity_w_Lineage):
    auth: dmda.DomoAuth = field(repr=False)
    id: str
    Lineage: dmdl.DomoLineage_Publication = field(repr=False)

    name: str
    description: str
    is_v2: bool
    created_dt: dt.datetime

    updated_dt: dt.datetime = None

    subscription_authorizations: List[DomoPublication_Subscription] = None

    content: List[DomoPublication_Content] = None

    # content_page_id_ls: List[str] = default = None
    # content_dataset_id_ls: List[str] = field(default_factory=list)
    # content_data_app_id_ls: List[str] = field(default_factory=list)


    def __post_init__(self):
        self.Lineage = dmdl.DomoLineage_Publication._from_parent(parent=self, auth=self.auth)

    def _generate_subscription_authorizations(
        self, subscription_authorizations_ls, auth
    ):
        self.subscription_authorizations = [
            DomoPublication_Subscription._from_json(obj=sub, auth=auth, parent=self)
            for sub in subscription_authorizations_ls
        ]

    def _generate_content(self, children_ls):

        self.content = [
            DomoPublication_Content._from_json(child, auth=self.auth, parent=self)
            for child in children_ls
        ]

        # if dmpc.entity_type == "PAGE":
        #     domo_pub.content_page_id_ls.append(dmpc.entity_id)

        # elif dmpc.entity_type == "DATASET":
        #     domo_pub.content_dataset_id_ls.append(dmpc.entity_id)

        # elif dmpc.entity_type == "DATA_APP":
        #     domo_pub.content_data_app_id_ls.append(dmpc.entity_id)

        # else:
        #     raise DomoPublication_UnexpectedContentType(
        #         cls_instance=domo_pub,
        #         content_type=dmpc.entity_type,
        #     )

        return self.content

    @classmethod
    def _from_json(cls, obj, auth: dmda.DomoAuth):

        domo_pub = cls(
            id=obj["id"],
            name=obj["name"],
            description=obj["description"],
            created_dt=(
                dt.datetime.fromtimestamp(obj["created"] / 1000)
                if obj["created"]
                else None
            ),
            updated_dt=(
                dt.datetime.fromtimestamp(obj.get("content").get("updated") / 1000)
                if obj.get("content").get("updated")
                else None
            ),
            is_v2=obj["isV2"],
            auth=auth,
            raw = obj,
            Lineage = None
        )

        if (
            obj.get("subscriptionAuthorizations")
            and len(obj.get("subscriptionAuthorizations")) > 0
        ):
            domo_pub._generate_subscription_authorizations(
                subscription_authorizations_ls=obj["subscriptionAuthorizations"],
                auth=auth,
            )

        if obj.get("children") and len(obj.get("children")) > 0:
            domo_pub._generate_content(obj["children"])

        return domo_pub
    
    @classmethod
    async def get_by_id(
        cls,
        publication_id,
        auth: dmda.DomoAuth = None,
        return_raw: bool = False,
        timeout=10,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        res = await publish_routes.get_publication_by_id(
            auth=auth,
            publication_id=publication_id,
            timeout=timeout,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        return cls._from_json(obj=res.response, auth=auth)
    
    @classmethod
    async def _get_entity_by_id(cls, entity_id, **kwargs):
        return await cls.get_by_id( publication_id = entity_id, **kwargs)
    
    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/admin/domo-everywhere/publications/details?id={self.id}"

# %% ../../nbs/classes/50_DomoPublish.ipynb 13
@patch_to(DomoPublication)
async def get_content(
    self,
    return_raw: bool = False,
    timeout=10,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    res = await publish_routes.get_publication_by_id(
        auth=self.auth,
        publication_id=self.id,
        timeout=timeout,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )

    content = res.response.get('children')

    if return_raw:
        res.response = content
        return res
    

    if not content or len(content) == 0:
        return content
    
    return self._generate_content(content)


# %% ../../nbs/classes/50_DomoPublish.ipynb 21
@dataclass
class DomoPublications:
    auth: dmda.DomoAuth = field(repr = False)

    publications: List[DomoPublication] = field(default = None)

    subscriptions: List[DomoPublication_Subscription] = field(default = None)

    async def get(
        self,
        search_term: str = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_num_stacks_to_drop=2,
    ):

        res = await publish_routes.search_publications(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            search_term= search_term,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
        )

        if return_raw:
            return res

        self.publications = await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoPublication.get_by_id(
                    publication_id=obj.get("id"), auth=self.auth
                )
                for obj in res.response
            ],
        )
        return self.publications

    async def get_subscribers(
        self,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_api: bool = False,
    ):
        """get instances subscription summaries"""

        res = await publish_routes.get_subscription_summaries(
            auth=self.auth, session=session, debug_api=debug_api
        )

        if return_raw:
            return res

        self.subscriptions = [
            DomoPublication_Subscription._from_json(sub, auth=self.auth)
            for sub in res.response
        ]

        return self.subscriptions

# %% ../../nbs/classes/50_DomoPublish.ipynb 25
@patch_to(DomoPublications)
async def search_publications(
    self: DomoPublications,
    search_term: str = None,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    return_raw: bool = False,
    debug_num_stacks_to_drop = 2
):
    
    res = await self.get(
        search_term=search_term,
        session=session,
        debug_api=debug_api,
        return_raw = return_raw,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop+1
    )

    return res

# %% ../../nbs/classes/50_DomoPublish.ipynb 27
@patch_to(DomoPublication)
async def report_content_as_dataframe(self, return_raw: bool = False):

    if return_raw:
        return await self.get_content()

    return pd.DataFrame(
        [
            {
                "plubication_id": self.id,
                "publication_name": self.name,
                "is_v2": self.is_v2,
                "publish_created_dt": self.created_dt,
                "entity_type": content.entity_type,
                "entity_id": content.entity_id,
                "entity_updated" : content.updated_dt
            }
            for content in self.content
        ]
    )


@patch_to(DomoPublication)
def report_lineage_as_dataframe(self, return_raw: bool = False):

    flat_lineage_ls = self.Lineage._flatten_lineage()

    output_ls = [
        {
            "plubication_id": self.id,
            "publication_name": self.name,
            "is_v2": self.is_v2,
            "publish_created_dt": self.created_dt,
            # "entity_type": row.get("entity_type"),
            "entity_id": row.get("entity_id"),
        }
        for row in flat_lineage_ls
    ]

    if return_raw:
        return output_ls

    return pd.DataFrame(output_ls)

# %% ../../nbs/classes/50_DomoPublish.ipynb 29
@patch_to(DomoPublication, cls_method=True)
async def create_publication(
    cls,
    name: str,
    content_ls: List[DomoPublication_Content],
    subscription_ls: List[DomoPublication_Subscription],
    unique_id: str = None,
    description: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
):

    if not isinstance(subscription_ls, list):
        subscription_ls = [subscription_ls]

    auth = auth or cls.auth
    domain_ls = []
    content_json_ls = []
    for sub in subscription_ls:
        domain_ls.append(sub.domain)
    for content_item in content_ls:
        content_json_ls.append(content_item.to_api_json())

    if not unique_id:
        unique_id = str(uuid.uuid4())
    if not description:
        description = ""

    body = publish_routes.generate_publish_body(
        url=f"{auth.domo_instance}.domo.com",
        sub_domain_ls=domain_ls,
        content_ls=content_json_ls,
        name=name,
        unique_id=unique_id,
        description=description,
        is_new=True,
    )

    res = await publish_routes.create_publish_job(auth=auth, body=body)
    if debug_api:
        print("Create the new Publish job")
    if res.status != 200:
        print(res)
        await asyncio.sleep(2)
        res = await publish_routes.get_publication_by_id(
            auth=auth, publication_id=unique_id
        )
        if res.status != 200:
            return None
        else:
            return cls._from_json(obj=res.response, auth=auth)

    return cls._from_json(obj=res.response, auth=auth)

# %% ../../nbs/classes/50_DomoPublish.ipynb 31
@patch_to(DomoPublication)
async def update_publication(
    self,
    name: str,
    content_ls: List[DomoPublication_Content],
    subscription_ls: List[DomoPublication_Subscription],
    description: str = "",
    debug_api: bool = False,
):

    if not isinstance(subscription_ls, list):
        subscription_ls = [subscription_ls]

    domain_ls = [sub.domain for sub in subscription_ls]
    content_json_ls = [content_item.to_api_json() for content_item in content_ls]
        
    body = publish_routes.generate_publish_body(
        url=f"{self.auth.domo_instance}.domo.com",
        sub_domain_ls=domain_ls,
        content_ls=content_json_ls,
        name=name,
        unique_id=self.id,
        description=description,
        is_new=False,
    )

    res = await publish_routes.udpate_publish_job(
        auth=self.auth, publication_id=self.id, body=body
    )

    return res

# %% ../../nbs/classes/50_DomoPublish.ipynb 33
@patch_to(DomoPublication, cls_method=True)
async def get_subscription_invites_list(
    cls, auth: dmda.DomoAuth, debug_api: bool = False
):

    res = await publish_routes.get_subscription_invititations(
        auth=auth, debug_api=debug_api
    )
    if debug_api:
        print("Getting Publish subscription invites")

    if res.status == 200:
        return res.response
    else:
        return None


@patch_to(DomoPublication, cls_method=True)
async def accept_invite_by_id(
    cls, auth: dmda.DomoAuth, subscription_id: str, debug_api: bool = False
):

    res = await publish_routes.accept_invite_by_id(
        auth=auth, subscription_id=subscription_id, debug_api=debug_api
    )

    if res.status == 200:
        return res.response
    else:
        return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/utils/compare.ipynb.

# %% auto 0
__all__ = ['compare_dicts']

# %% ../../nbs/utils/compare.ipynb 3
def compare_dicts(dict1, dict2, path=""):
    """
    Recursively compare two dictionaries (and lists) and print differences.
    """

    diff_ls = []
    if isinstance(dict1, dict) and isinstance(dict2, dict):
        for key in dict1.keys():
            if key not in dict2:
                if dict1[key] is not None:
                    diff_ls.append({"key" : path + str(key), "message": 'found in first dict but not in second.'})
            else:
                compare_dicts(dict1[key], dict2[key], path + f"{key}.")
        for key in dict2.keys():
            if key not in dict1:
                if dict2[key] is not None:
                    diff_ls.append({"key" : path + str(key), "message": 'found in second dict but not in first.'})
    elif isinstance(dict1, list) and isinstance(dict2, list):
        min_len = min(len(dict1), len(dict2))
        for i in range(min_len):
            compare_dicts(dict1[i], dict2[i], path + f"[{i}].")
        if len(dict1) > len(dict2):
            for i in range(min_len, len(dict1)):
                diff_ls.append({"key" : f"{path}[{i}]", "message": f'Extra item in first list at \'{path}[{i}]\': {dict1[i]}'})
        elif len(dict2) > len(dict1):
            for i in range(min_len, len(dict2)):
                diff_ls.append({"key" : f"{path}[{i}]", "message": f'Extra item in second list at \'{path}[{i}]\': {dict2[i]}'})
    else:
        if dict1 != dict2:
            diff_ls.append(f"Value mismatch at '{path[:-1]}': {dict1} != {dict2}")

    return diff_ls



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb.

# %% auto 0
__all__ = ['Watchdog_Config', 'Watchdog_Config_MaxIndexingTime', 'Watchdog_Config__Variance', 'Watchdog_Config_RowCountVariance',
           'Watchdog_Config_ExecutionVariance', 'Watchdog_Config_ErrorDetection', 'Watchdog_Config_LastDataUpdated',
           'Watchdog_Config_CustomQuery', 'Watchdog_ConfigFactory', 'DomoJob_Watchdog']

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 3
from dataclasses import dataclass, field
from abc import ABC, abstractmethod

from typing import List, Optional
from enum import Enum

import httpx

from nbdev.showdoc import patch_to

import domolibrary.client.DomoAuth as dmda
import datetime as dt

from pprint import pprint

import domolibrary.routes.application as application_routes


from domolibrary.classes.DomoApplication_Job_Base import DomoJob_Base, DomoTrigger_Schedule

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 7
@dataclass
class Watchdog_Config(ABC):
    entity_ids: List[str]
    entity_type: str  # dataflow or dataset
    watcher_parameters: dict
    report_type: str

    """base class for the different watchdog report types"""

    @classmethod
    def _from_json(cls, watcher_parameters_obj):  # executionPayload

        return cls(
            entity_ids=watcher_parameters_obj["entityIds"],
            entity_type=watcher_parameters_obj["entityType"],
            report_type=watcher_parameters_obj["type"],
            watcher_parameters=watcher_parameters_obj,
        )

    def _to_json(self):
        return {
            "entityIds": self.entity_ids,
            "entityType": self.entity_type,
            "type": self.report_type,
        }

    @abstractmethod
    def to_json(self):
        return self._to_json()

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 8
@dataclass
class Watchdog_Config_MaxIndexingTime(Watchdog_Config):
    report_type = "max_indexing_time"
    max_indexing_time_mins = 30

    def __post_init__(self):
        self.max_indexing_time_mins = self.watcher_parameters[
            "maxIndexingTimeInMinutes"
        ]

    def to_json(self, **kwargs):
        if kwargs.get("max_indexing_time_mins"):
            self.max_indexing_time_mins = kwargs["max_indexing_time_mins"]

        return {
            **self._to_json(),
            "maxIndexingTimeInMinutes": self.max_indexing_time_mins,
        }


@dataclass
class Watchdog_Config__Variance(Watchdog_Config):
    report_type: str
    variance_percent: int = 10

    def __post_init__(self):
        self.variance_percent = self.watcher_parameters["variancePercent"]

    def to_json(self, **kwargs):
        if kwargs.get("variance_percent"):
            self.variance_percent = kwargs["variance_percent"]

        return {
            **self._to_json(),
            "variancePercent": self.variance_percent,
        }


@dataclass
class Watchdog_Config_RowCountVariance(Watchdog_Config__Variance):
    report_type = "row_count_variance"


@dataclass
class Watchdog_Config_ExecutionVariance(Watchdog_Config__Variance):
    report_type = "execution_variance"


@dataclass
class Watchdog_Config_ErrorDetection(Watchdog_Config):
    report_type: str = "error_detection"

    def to_json(self, **kwargs):
        return {
            **self._to_json(),
        }


@dataclass
class Watchdog_Config_LastDataUpdated(Watchdog_Config):
    report_type: str = "last_data_updated"
    min_data_update_frequency_in_mins: int = 10

    def __post_init__(self):
        self.min_data_update_frequency_in_mins = self.watcher_parameters[
            "minDataUpdateFrequencyInMinutes"
        ]

    def to_json(self, **kwargs):
        if kwargs.get("min_data_update_frequency_in_mins"):
            self.min_data_update_frequency_in_mins = kwargs[
                "min_data_update_frequency_in_mins"
            ]

        return {
            **self._to_json(),
            "minDataUpdateFrequencyInMinutes": self.min_data_update_frequency_in_mins,
        }


@dataclass
class Watchdog_Config_CustomQuery(Watchdog_Config):
    report_type: str = "custom_query"
    sql_query: str = ""

    def __post_init__(self):
        self.sql_query = self.watcher_parameters["sqlQuery"]

    def to_json(self, **kwargs):
        if kwargs.get("sql_query"):
            self.sql_query = kwargs["sql_query"]

        return {
            **self._to_json(),
            "sqlQuery": self.sql_query,
        }

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 9
class Watchdog_ConfigFactory(Enum):
    MAX_INDEXING_TIME = Watchdog_Config_MaxIndexingTime
    ROW_COUNT_VARIANCE = Watchdog_Config_RowCountVariance
    EXECUTION_VARIANCE = Watchdog_Config_ExecutionVariance
    ERROR_DETECTION = Watchdog_Config_ErrorDetection
    LAST_DATA_UPDATED = Watchdog_Config_LastDataUpdated
    CUSTOM_QUERY = Watchdog_Config_CustomQuery

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 11
@dataclass
class DomoJob_Watchdog(DomoJob_Base):
    custom_message: str = None
    remote_instance: str = None

    notify_emails: List[str] = field(default_factory=lambda: [])
    notify_group_ids: List[str] = field(default_factory=lambda: [])
    notify_user_ids: List[str] = field(default_factory=lambda: [])

    Config: Watchdog_Config = None
    webhooks: List[str] = None

    @classmethod
    def _from_json(cls, obj, auth):

        remote_instance = cls._format_remote_instance(
            obj["executionPayload"].get("domain")
        )

        watchdog_parameters_obj = obj["executionPayload"]["watcherParameters"]

        config = Watchdog_ConfigFactory[
            watchdog_parameters_obj["type"].upper()
        ].value._from_json(watchdog_parameters_obj)

        return cls(
            **cls._convert_API_res_to_DomoJob_base_obj(obj),
            remote_instance=remote_instance,
            custom_message=obj["executionPayload"]["customMessage"],
            Config=config,
            notify_emails=obj["executionPayload"]["notifyEmailAddresses"],
            notify_group_ids=obj["executionPayload"]["notifyGroupIds"],
            notify_user_ids=obj["executionPayload"]["notifyUserIds"],
            webhooks=obj["executionPayload"]["webhooks"],
            auth=auth
        )

    @classmethod
    async def get_by_id(
        cls,
        application_id,
        job_id,
        auth: dmda.DomoAuth,
        debug_api: bool = False,
        session: Optional[httpx.AsyncClient] = None,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False,
    ):

        return await cls._get_by_id(
            application_id=application_id,
            job_id=job_id,
            auth=auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
            new_cls=cls,
            parent_class=cls.__name__,
        )

    def to_json(self):
        s = self._generate_to_json()

        s["executionPayload"].update(
            {
                "customMessage": self.custom_message,
                "domain": self.remote_instance,
                "notifyEmailAddresses": self.notify_emails,
                "notifyGroupIds": self.notify_group_ids,
                "notifyUserIds": self.notify_user_ids,
                "watcherParameters": self.Config.to_json(),
                "webhooks": self.webhooks,
            }
        )
        return s

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 14
@patch_to(DomoJob_Base)
async def update(
    self: DomoJob_Base,
    debug_api: bool = False,
    session: Optional[httpx.AsyncClient] = None,
    debug_num_stacks_to_drop=2,
):

    res = await application_routes.update_application_job(
        auth=self.auth,
        body=self.to_json(),
        application_id=self.application_id,
        job_id=self.id,
        debug_api=debug_api,
        session=session,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    return res

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 17
@patch_to(DomoJob_Base)
async def execute(
    self,
    debug_api: bool = False,
    session: Optional[httpx.AsyncClient] = None,
    debug_num_stacks_to_drop=2,
):

    res = await application_routes.execute_application_job(
        auth=self.auth,
        application_id=self.application_id,
        job_id=self.id,
        debug_api=debug_api,
        session=session,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    return res

# %% ../../nbs/classes/50_DomoApplication_Job_Watchdog.ipynb 20
@patch_to(DomoJob_Watchdog, cls_method=True)
async def create(
    cls,
    auth: dmda.DomoAuth,
    name: str,
    application_id: str,
    config: Watchdog_Config,
    logs_dataset_id: str,
    notify_user_ids: list = None,
    notify_group_ids: list = None,
    notify_emails: list = None,
    triggers: List[DomoTrigger_Schedule] = None,
    description: str = f"created via domolibrary - {dt.date.today()}",
    execution_timeout=1440,
    accounts: list[int] = None,
    remote_instance=None,
    custom_message: str = None,
    webhooks: List[str] = None,
    debug_api: bool = False,
    session: Optional[httpx.AsyncClient] = None,
    return_raw: bool = False,
):

    domo_job = cls(
        auth=auth,
        name=name,
        description=description,
        application_id=application_id,
        execution_timeout=execution_timeout,
        notify_user_ids=notify_user_ids or [],
        notify_group_ids=notify_group_ids or [],
        notify_emails=notify_emails or [],
        triggers=triggers or [],
        accounts=accounts or [],
        logs_dataset_id=logs_dataset_id,
        remote_instance=remote_instance,
        Config=config,
        custom_message=custom_message,
        webhooks=webhooks,
    )

    body = domo_job.to_json()

    res = await application_routes.create_application_job(
        auth=auth,
        application_id=application_id,
        body=body,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    return cls._from_json(res.response, auth=auth)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoPage.ipynb.

# %% auto 0
__all__ = ['DomoPage_GetRecursive', 'DomoPage', 'DomoPages', 'Page_NoAccess']

# %% ../../nbs/classes/50_DomoPage.ipynb 2
from ..routes.page import PageRetrieval_byId_Error,Page_GET_Error,CRUD_Page_Error

# %% ../../nbs/classes/50_DomoPage.ipynb 3
from dataclasses import dataclass, field
from typing import List, Any

import httpx
import datetime as dt


import domolibrary.client.DomoError as dmde
import domolibrary.client.DomoAuth as dmda
from domolibrary.client.DomoEntity import DomoEntity_w_Lineage

import domolibrary.routes.page as page_routes

import domolibrary.utils.convert as dmcv
import domolibrary.utils.DictDot as util_dd
import domolibrary.utils.chunk_execution as dmce

import domolibrary.classes.DomoPage_Content as dmpg_c
import domolibrary.classes.DomoLineage as dmdl

from nbdev.showdoc import patch_to


# %% ../../nbs/classes/50_DomoPage.ipynb 7
class DomoPage_GetRecursive(dmde.ClassError):
    def __init__(
        self,
        cls,
        entity_id,
        auth: dmda.DomoAuth,

        include_recursive_children,
        include_recursive_parents,
    ):
        super().__init__(
            auth = auth,
            cls = cls,
            entity_id=entity_id,
            message=f"can only trace parents OR children recursively but not both. include_recursive_children : {include_recursive_children}, include_recursive_parents: {include_recursive_parents}",
        )


@dataclass
class DomoPage(DomoEntity_w_Lineage):
    id: int
    auth: dmda.DomoAuth = field(repr=False)
    Lineage: dmdl.DomoLineage = field(repr=False)
    
    title: str = None
    top_page_id: int = None
    parent_page_id: int = None
    is_locked: bool = None

    collections: list = field(default_factory=list)

    owners: list = field(default_factory=list)
    cards: list = field(default_factory=list)

    custom_attributes: dict = field(default_factory=dict)

    parent_page: dict = None  # DomoPage
    top_page: dict = None  # DomoPage
    children: list = field(default_factory=list)

    # parent_hierarchy: [dict] = None
    # flat_children: list = None

    layout: dmpg_c.PageLayout = field(default_factory=dict)

    cards: List["DomoCard"] = None
    datasets: List["DomoDataset"] = None


    def __post_init__(self):
        self.Lineage = dmdl.DomoLineage_Page._from_parent(
            parent=self
        )

    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/page/{self.id}"

    async def _get_domo_owners_from_dd(
        self,
        owners: util_dd.DictDot,
        suppress_no_results_error: bool = True,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 3,
    ):
        if not owners or len(owners) == 0:
            return []

        import domolibrary.classes.DomoUser as dmu
        import domolibrary.classes.DomoGroup as dmg

        domo_groups = []
        domo_users = []

        owner_group_ls = [
            owner.id for owner in owners if owner.type == "GROUP" and owner.id
        ]

        if len(owner_group_ls) > 0:
            domo_groups = await dmce.gather_with_concurrency(
                n=60,
                *[
                    dmg.DomoGroup.get_by_id(
                        group_id=group_id,
                        auth=self.auth,
                        debug_api=debug_api,
                        session=session,
                        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                    )
                    for group_id in owner_group_ls
                ],
            )

        owner_user_ls = [
            owner.id for owner in owners if owner.type == "USER" and owner.id
        ]

        if len(owner_user_ls) > 0:
            domo_users = await dmu.DomoUsers.by_id(
                user_ids=owner_user_ls,
                only_allow_one=False,
                auth=self.auth,
                session=session,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
                suppress_no_results_error=suppress_no_results_error,
            )

        owner_ce = (domo_groups or []) + (domo_users or [])

        res = []
        for owner in owner_ce:
            if isinstance(owner, list):
                [res.append(member) for member in owner]
            else:
                res.append(owner)

        return res
    
        
    @classmethod
    async def _from_content_stacks_v3(
        cls,
        page_obj,
        suppress_no_results_error: bool = True,
        auth: dmda.DomoAuth = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 3,
    ):
        # import domolibrary.classes.DomoCard as dc

        dd = page_obj
        if isinstance(page_obj, dict):
            dd = util_dd.DictDot(page_obj)

        pg = cls(
            id=int(dd.id),
            title=dd.title,
            raw = page_obj,
            parent_page_id=int(dd.page.parentPageId) if dd.page.parentPageId else None,
            collections=dd.collections,
            auth=auth,
            Lineage = None,
        )

        if hasattr(dd, "pageLayoutV4") and dd.pageLayoutV4 is not None:
            pg.layout = dmpg_c.PageLayout._from_json(dd=dd.pageLayoutV4)

        if dd.page.owners and len(dd.page.owners) > 0:
            pg.owners = await pg._get_domo_owners_from_dd(
                dd.page.owners,
                suppress_no_results_error=suppress_no_results_error,
                debug_api=debug_api,
                session=session,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )

        return pg
    

    @classmethod
    async def get_by_id(
        cls,
        page_id: str,
        auth: dmda.DomoAuth,
        suppress_no_results_error : bool = True,
        return_raw: bool = False,
        debug_api: bool = False,
        include_layout: bool = False,
        # if True, will drill down to all the Children.  Set to False to prevent calculating children
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
        id=None,
    ):
        page_id = page_id or id

        res = await page_routes.get_page_by_id(
            auth=auth,
            page_id=page_id,
            debug_api=debug_api,
            include_layout=include_layout,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res


        pg = await cls._from_content_stacks_v3(
            page_obj=res.response,
            auth=auth,
            session=session,
            debug_api=debug_api,
            suppress_no_results_error=suppress_no_results_error,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        pg.custom_attributes["parent_page"] = None
        pg.custom_attributes["top_page"] = None

        return pg

    @classmethod
    async def _get_entity_by_id(
        cls,
        entity_id: str,
        **kwargs
    ):
        return await cls.get_by_id(
            page_id=entity_id,
            **kwargs
        )

# %% ../../nbs/classes/50_DomoPage.ipynb 9
@patch_to(DomoPage, cls_method=True)
async def _from_adminsummary(
    cls,
    page_obj,
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
):
    import domolibrary.classes.DomoCard as dmc

    dd = page_obj

    if isinstance(page_obj, dict):
        dd = util_dd.DictDot(page_obj)

    page_id = int(dd.id or dd.pageId)

    parent_page_id = int(dd.parentPageId) if dd.parentPageId else page_id

    top_page_id = int(dd.topPageId) if dd.topPageId else parent_page_id

    pg = cls(
        id=page_id,
        title=dd.title or dd.pageTitle,
        raw=page_obj,
        parent_page_id=parent_page_id,
        top_page_id=top_page_id,
        collections=dd.collections,
        is_locked=dd.locked,
        auth=auth,
        Lineage=None,
    )

    if dd.page and dd.page.owners and len(dd.page.owners) > 0:
        pg.owners = await pg._get_domo_owners_from_dd(
            dd.page.owners, debug_api=debug_api, session=session
        )

    if dd.cards and len(dd.cards) > 0:
        pg.cards = await dmce.gather_with_concurrency(
            n=60,
            *[
                dmc.DomoCard.get_by_id(
                    id=card.id, auth=auth, session=session, debug_api=debug_api
                )
                for card in dd.cards
            ],
        )

    return pg


@patch_to(DomoPage, cls_method=True)
async def _from_bootstrap(
    cls: DomoPage,
    page_obj,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    dd = page_obj
    if isinstance(page_obj, dict):
        dd = util_dd.DictDot(page_obj)

    pg = cls(id=int(dd.id), title=dd.title, raw=page_obj, auth=auth, Lineage = None)

    if isinstance(dd.owners, list) and len(dd.owners) > 0:
        pg.owners = await pg._get_domo_owners_from_dd(
            dd.owners, debug_api=debug_api, session=session
        )

    if isinstance(dd.children, list) and len(dd.children) > 0:
        pg.children = await dmce.gather_with_concurrency(
            n=60,
            *[
                cls._from_bootstrap(
                    page_obj=child_dd, auth=auth, debug_api=debug_api, session=session
                )
                for child_dd in dd.children
                if child_dd.type == "page"
            ],
        )

        [print(other_dd) for other_dd in dd.children if other_dd.type != "page"]

    return pg

# %% ../../nbs/classes/50_DomoPage.ipynb 11
@dataclass
class DomoPages:
    auth: dmda.DomoAuth = field(repr=False)
    pages: List[DomoPage] = None

    async def get(self, **kwargs):
        """calls get_admin_summary to retrieve all pages in an instance"""
        return await self.get_admin_summary(**kwargs)

    async def get_admin_summary(
        self,
        search_title: str = None,
        parent_page_id: int = None,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        """use admin_summary to retrieve all pages in an instance -- regardless of user access
        NOTE: some Page APIs will not return results if page access isn't explicitly shared
        """

        res = await page_routes.get_pages_adminsummary(
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            search_title=search_title,
            page_parent_id=parent_page_id,
        )

        if return_raw:
            return res

        self.pages = await dmce.gather_with_concurrency(
            n=60,
            *[
                DomoPage._from_adminsummary(
                    page_obj, auth=self.auth, debug_api=debug_api, session=session
                )
                for page_obj in res.response
            ],
        )

        return self.pages

# %% ../../nbs/classes/50_DomoPage.ipynb 15
@patch_to(DomoPage)
async def get_children(self: DomoPage, is_suppress_errors: bool = False):

    async def _get_children_recur(parent_page, is_suppress_errors: bool = False):

        parent_page.children = parent_page.children or []

        try:
            child_pages = await DomoPages(auth=parent_page.auth).get(
                parent_page_id=parent_page.id, 
                # is_suppress_errors=is_suppress_errors
            )

            parent_page.children = [
                child
                for child in child_pages
                if child is not None and child.parent_page_id == parent_page.id
            ]

            await dmce.gather_with_concurrency(
                n=10,
                *[
                    _get_children_recur(
                        parent_page=cp,
                        is_suppress_errors=is_suppress_errors,
                    )
                    for cp in parent_page.children
                ],
            )

            return self.children

        except dmde.DomoError as e:
            print(
                f"cannot access child page -- https://{parent_page.auth.domo_instance}.domo.com/page/{page_id} -- is it shared with you?"
            )
            if not is_suppress_errors:
                raise e from e

    self.children = await _get_children_recur(
        parent_page=self,
        is_suppress_errors=is_suppress_errors,
    )

    return self.children


@patch_to(DomoPage)
def flatten_children(self: DomoPage, path=None, hierarchy=0, results=None):
    results = results or []

    path = f"{path} > {self.title}" if path else self.title

    results.append({"hierarchy": hierarchy, "path": path, "page": self})

    if self.children:
        [
            child.flatten_children(path, hierarchy + 1, results)
            for child in self.children
        ]

    return results

# %% ../../nbs/classes/50_DomoPage.ipynb 18
@patch_to(DomoPage)
async def get_parents(self: DomoPage,page = None):

    page = page or self

    if not page.parent_page_id:
        return self.custom_attributes
    
    if not self.top_page_id: 
        page_as = next((pg for pg in (await DomoPages(auth = self.auth).get(search_title=page.id)) if pg.id ==page.id))
        page.top_page_id =  page_as.top_page_id
        self.top_page = page_as
        
    if page.id == page.top_page_id:
        self.custom_attributes['top_page'] = page

    parent_page_as = next((pg for pg in (await DomoPages(auth = self.auth).get(search_title=page.parent_page_id)) if pg.id == page.parent_page_id))

    if self.parent_page_id == parent_page_as.id:
        self.parent_page = parent_page_as

    page.custom_attributes["parent_page"] = parent_page_as
    
    if not self.custom_attributes.get('path'):
        self.custom_attributes["path"] = []

    self.custom_attributes['path'].append(parent_page_as)

    if page.id != self.top_page_id:
        await self.get_parents(page = parent_page_as)

    return self.custom_attributes



# %% ../../nbs/classes/50_DomoPage.ipynb 20
class Page_NoAccess(dmde.DomoError):
    def __init__(self, page_id, page_title, domo_instance, function_name, parent_class):
        super().__init__(
            function_name=function_name,
            parent_class=parent_class,
            domo_instance=domo_instance,
            message=f'authenticated user doesn\'t have access to {page_id} - "{page_title}" contact owners to share access',
        )

# %% ../../nbs/classes/50_DomoPage.ipynb 21
@patch_to(DomoPage)
async def test_page_access(
    self: DomoPage,
    suppress_no_access_error: bool = False,  # suppresses error if user doesn't have access
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
):
    """throws an error if user doesn't have access to the page
    API returns the owners of the page
    """

    res = await page_routes.get_page_access_test(
        auth=self.auth,
        page_id=self.id,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class= self.__class__.__name__
    )

    if return_raw:
        return res

    try:
        page_access = res.response.get("pageAccess")

        if not page_access:
            raise Page_NoAccess(
                page_id=self.id,
                page_title=self.title,
                domo_instance=self.auth.domo_instance,
                function_name=res.traceback_details.function_name,
                parent_class=self.__class__.__name__,
            )

    except Page_NoAccess as e:
        print(e)

        if not suppress_no_access_error:
            raise e

    return res

# %% ../../nbs/classes/50_DomoPage.ipynb 25
@patch_to(DomoPage)
async def get_accesslist(
    self,
    auth: dmda.DomoAuth = None,
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
):
    auth = auth or self.auth

    res = await page_routes.get_page_access_list(
        auth=auth,
        is_expand_users=True,
        page_id=self.id,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res


    import domolibrary.classes.DomoUser as dmu
    import domolibrary.classes.DomoGroup as dmg

    s = {
        "explicit_shared_user_count": res.response.get("explicitSharedUserCount"),
        "total_user_count": res.response.get("totalUserCount"),
    }

    user_ls = res.response.get("users", None)
    domo_users = []
    
    if user_ls and isinstance(user_ls, list) and len(user_ls) > 0:
        domo_users = await dmu.DomoUsers.by_id(
            user_ids=[user.get("id") for user in user_ls],
            only_allow_one=False,
            auth=auth,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

    group_ls = res.response.get("groups", None)
    domo_groups = []
    if group_ls and isinstance(group_ls, list) and len(group_ls) > 0:
        domo_groups = await dmce.gather_with_concurrency(
            n=60,
            *[
                dmg.DomoGroup.get_by_id(
                    group_id=group.get("id"),
                    auth=auth,
                    session=session,
                    debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
                )
                for group in group_ls
            ],
        )

    res = await self.test_page_access(suppress_no_access_error=True)
    owner_ls = res.response["owners"]  # from test_page_access

    for domo_user in domo_users:
        # isExplicitShare is set by the get_access_list API response
        domo_user.custom_attributes["is_explicit_share"] = next(
            (
                user_obj["isExplicitShare"]
                for user_obj in user_ls
                if int(user_obj.get("id")) == int(domo_user.id)
            )
        )

        # group membership is determined by get_access_list API response
        domo_user.custom_attributes["group_membership"] = [
            domo_group
            for group_obj in group_ls
            for domo_group in domo_groups
            if int(domo_user.id)
            in [int(user_obj["id"]) for user_obj in group_obj.get("users")]
            and domo_group.id == group_obj["id"]
        ]

        # isOwner determined by test_access API response and group membership
        domo_user.custom_attributes["is_owner"] = False

        # test ownership as a user
        match_owner = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"]) == int(domo_user.id)
                and owner_obj["type"] == "USER"
            ),
            None,
        )

        match_group = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"])
                in [
                    int(domo_group.id)
                    for domo_group in domo_user.custom_attributes["group_membership"]
                ]
                and owner_obj["type"] == "GROUP"
            ),
            None,
        )

        if match_owner or match_group:
            domo_user.custom_attributes["is_owner"] = True

    # group ownership is confirmed test_access API
    for domo_group in domo_groups:
        match_owner = next(
            (
                owner_obj
                for owner_obj in owner_ls
                if int(owner_obj["id"]) == int(domo_group.id)
                and owner_obj["type"] == "GROUP"
            ),
            None,
        )

        domo_group.custom_attributes["is_owner"] = True if match_owner else False

    return {
        **s,
        "domo_users": domo_users,
        "domo_groups": domo_groups,
    }

# %% ../../nbs/classes/50_DomoPage.ipynb 28
@patch_to(DomoPage)
async def share(
    self: DomoPage,
    auth: dmda.DomoAuth = None,
    domo_users: list = None,  # DomoUsers to share page with,
    domo_groups: list = None,  # DomoGroups to share page with
    message: str = None,  # message for automated email
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    import domolibrary.routes.datacenter as datacenter_routes

    if domo_groups:
        domo_groups = domo_groups if isinstance(domo_groups, list) else [domo_groups]
    if domo_users:
        domo_users = domo_users if isinstance(domo_users, list) else [domo_users]

    res = await datacenter_routes.share_resource(
        auth=auth or self.auth,
        resource_ids=[self.id],
        resource_type=datacenter_routes.ShareResource_Enum.PAGE,
        group_ids=[group.id for group in domo_groups] if domo_groups else None,
        user_ids=[user.id for user in domo_users] if domo_users else None,
        message=message,
        debug_api=debug_api,
        session=session,
    )

    return res

# %% ../../nbs/classes/50_DomoPage.ipynb 31
@patch_to(DomoPage)
async def get_cards(
    self,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw : bool = False
):
    import domolibrary.classes.DomoCard as dc

    res = await page_routes.get_page_definition(
        auth=self.auth, page_id=self.id, debug_api=debug_api, session=session
    )

    if return_raw: 
        return res

    if len(res.response.get("cards")) == 0:
        return []

    self.domo_cards = await dmce.gather_with_concurrency(
        n=60,
        *[
            dc.DomoCard.get_by_id(card_id=card["id"], auth=self.auth)
            for card in res.response.get("cards")
        ],
    )

    return self.domo_cards

@patch_to(DomoPage)
async def get_datasets(
    self,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False
):
    import domolibrary.classes.DomoDataset as dmds

    res = await page_routes.get_page_definition(
        auth=self.auth, page_id=self.id, debug_api=debug_api, session=session
    )

    if return_raw:
        return res

    cards = await self.get_cards()

    card_datasets = await dmce.gather_with_concurrency(*[
        card.get_datasets(debug_api = debug_api, session = session) for card in cards], n = 10)

    self.datasets = [ds for ds_ls in card_datasets for ds in ds_ls if ds is not None]
    
    return self.datasets

# %% ../../nbs/classes/50_DomoPage.ipynb 34
@patch_to(DomoPage, cls_method=True)
async def update_layout(
    cls, auth: dmda.DomoAuth, body: dict, layout_id: str, debug_api: bool = False
):
    datetime_now = dt.datetime.now()
    start_time_epoch = dmcv.convert_datetime_to_epoch_millisecond(datetime_now)

    res_writelock = await page_routes.put_writelock(
        auth=auth,
        layout_id=layout_id,
        user_id=auth.user_id,
        epoch_time=start_time_epoch,
    )
    if res_writelock.status == 200:
        res = await page_routes.update_page_layout(
            auth=auth, body=body, layout_id=layout_id, debug_api=debug_api
        )

        if not res.is_success:
            return False

        res_writelock = await page_routes.delete_writelock(
            auth=auth, layout_id=layout_id
        )
        if res_writelock.status != 200:
            return False

    else:
        return False

    return True

# %% ../../nbs/classes/50_DomoPage.ipynb 37
@patch_to(DomoPage,)
async def add_owner(
    self,
    group_id_ls: List[int],  # DomoGroup IDs to share page with
    user_id_ls: List[int],  # DomoUser IDs to share page with
    note: str = None,  # message for automated email
    send_email: bool = False,  # send or not email to the new owners
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    res = await page_routes.add_page_owner(
        auth=self.auth,
        page_id_ls=[self.id],
        group_id_ls=group_id_ls,
        user_id_ls=user_id_ls,
        note=note,
        send_email=send_email,
        debug_api=debug_api,
        session=session,
    )

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoApplication_Job.ipynb.

# %% auto 0
__all__ = []

# %% ../../nbs/classes/50_DomoApplication_Job.ipynb 3
from domolibrary.classes.DomoApplication_Job_Base import (
    DomoJob_Base as DomoJob,
    DomoTrigger,
    DomoTrigger_Schedule,
)
from domolibrary.classes.DomoApplication_Job_RemoteDomoStats import (
    DomoJob_RemoteDomoStats,
    RemoteDomoStats_Config,
    RemoteDomoStats_Config_Policy,
)
from domolibrary.classes.DomoApplication_Job_Watchdog import (
    Watchdog_Config,
    Watchdog_ConfigFactory,
    DomoJob_Watchdog,
    Watchdog_Config_CustomQuery,
    Watchdog_Config_ErrorDetection,
    Watchdog_Config_ExecutionVariance,
    Watchdog_Config_LastDataUpdated,
    Watchdog_Config_MaxIndexingTime,
    Watchdog_Config_RowCountVariance,
)

# %% ../../nbs/classes/50_DomoApplication_Job.ipynb 4
import domolibrary.client.DomoAuth as dmda
import datetime as dt

from pprint import pprint

import domolibrary.routes.application as application_routes



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoInstanceConfig_MFA.ipynb.

# %% auto 0
__all__ = ['MFAConfig_InstantiationError', 'MFA_Config']

# %% ../../nbs/classes/50_DomoInstanceConfig_MFA.ipynb 2
import httpx
from nbdev.showdoc import patch_to
from typing import List

from dataclasses import dataclass, field
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde

import domolibrary.routes.instance_config_mfa as mfa_routes

# %% ../../nbs/classes/50_DomoInstanceConfig_MFA.ipynb 6
class MFAConfig_InstantiationError(dmde.ClassError):

    def __init__(self, message, auth: dmda.DomoAuth, cls):

        super().__init__(domo_instance=auth.domo_instance, message=message, cls=cls)


@dataclass
class MFA_Config:
    auth: dmda.DomoAuth = field(repr=False)
    is_multifactor_required: bool = None
    max_code_attempts: int = None
    num_days_valid: int = None

    @classmethod
    def _from_json(cls, auth: dmda.DomoAuth, obj: List[dict]):

        return cls(
            auth=auth,
            is_multifactor_required=obj.get("is_multifactor_required"),
            num_days_valid=obj.get("num_days_valid"),
            max_code_attempts=obj.get("max_code_attempts"),
        )

    @classmethod
    async def get_instance_config(
        cls,
        auth: dmda.DomoAuth,
        incl_is_multifactor_required: bool = True,
        incl_num_days_valid: bool = True,
        incl_max_code_attempts: bool = True,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):
        res = await mfa_routes.get_mfa_config(
            auth=auth,
            incl_is_multifactor_required=incl_is_multifactor_required,
            incl_num_days_valid=incl_num_days_valid,
            incl_max_code_attempts=incl_max_code_attempts,
            session=session,
            debug_api=debug_api,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
        )

        if return_raw:
            return res

        if res.is_success and not (isinstance(res.response, dict)):
            raise MFAConfig_InstantiationError(
                auth=auth,
                message=f"must enable MFA to get MFA config from {auth.domo_instance}",
            )

        return cls._from_json(auth=auth, obj=res.response)

    async def get(
        self,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        return_raw: bool = False
    ):
        res = await mfa_routes.get_mfa_config(
            auth=self.auth,
            incl_is_multifactor_required=True,
            incl_num_days_valid=True,
            incl_max_code_attempts=True,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.num_days_valid = res.response.get("num_days_valid")
        self.max_code_attempts = res.response.get("max_code_attempts")
        self.is_multifactor_required = res.response.get("is_multifactor_required")

        return self

# %% ../../nbs/classes/50_DomoInstanceConfig_MFA.ipynb 8
@patch_to(MFA_Config)
def _test_need_update_config(self, attr_name, new_value, is_ignore_test: bool = False, debug_prn: bool = False):
    current_value = getattr(self, attr_name)

    if new_value is None:
        return False

    if new_value == current_value and not is_ignore_test:
        if debug_prn:
            print(f"No change necessary -- {current_value} is the same as {new_value}")
        return False

    return True


@patch_to(MFA_Config)
async def toggle_mfa(
    self,
    is_enable_MFA: bool,
    is_ignore_test: bool = False,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    suppress_update_self: bool = False,
    debug_prn: bool = False
):
    need_update_config = self._test_need_update_config(
        attr_name="is_multifactor_required",
        new_value=is_enable_MFA,
        is_ignore_test=is_ignore_test,
        debug_prn = debug_prn
    )

    if not need_update_config:
        return self

    await mfa_routes.toggle_enable_mfa(
        auth=self.auth,
        is_enable_MFA=is_enable_MFA,
        session=session,
        debug_api=debug_api,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if suppress_update_self:
        return self

    return await self.get()


@patch_to(MFA_Config)
async def enable(
    self,
    is_ignore_test: bool = False,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
):
    """will have enable_value=True and will enable mfa in the instance"""
    return await self.toggle_mfa(
        is_ignore_test=is_ignore_test,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        is_enable_MFA=True,
    )


@patch_to(MFA_Config)
async def disable(
    self,
    is_ignore_test: bool = False,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
):
    """will have enable_value=False and will disable mfa in the instance"""
    return await self.toggle_mfa(
        is_ignore_test=is_ignore_test,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        is_enable_MFA=False,
    )


@patch_to(MFA_Config)
async def set_max_code_attempts(
    self,
    max_code_attempts: int,
    is_ignore_test: bool = False,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    suppress_update_self: bool = False,
    debug_prn: bool = False
):
    need_update_config = self._test_need_update_config(
        attr_name="max_code_attempts",
        new_value=max_code_attempts,
        is_ignore_test=is_ignore_test,
        debug_prn=debug_prn
    )

    if not need_update_config:
        return self

    await mfa_routes.set_mfa_max_code_attempts(
        auth=self.auth,
        max_code_attempts=max_code_attempts,
        session=session,
        debug_api=debug_api,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if suppress_update_self:
        return self

    return await self.get()


@patch_to(MFA_Config)
async def set_num_days_valid(
    self,
    num_days_valid: int,
    is_ignore_test: bool = False,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    suppress_update_self: bool = False,
    debug_prn: bool = False
):
    need_update_config = self._test_need_update_config(
        attr_name="num_days_valid",
        new_value=num_days_valid,
        is_ignore_test=is_ignore_test,
        debug_prn=debug_prn
    )

    if not need_update_config:
        return self

    await mfa_routes.set_mfa_num_days_valid(
        auth=self.auth,
        num_days_valid=num_days_valid,
        session=session,
        debug_api=debug_api,
        parent_class=self.__class__.__name__,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )
    if not suppress_update_self:
        return self

    return await self.get()

# %% ../../nbs/classes/50_DomoInstanceConfig_MFA.ipynb 10
@patch_to(MFA_Config)
async def update(
    self,
    is_enable_MFA: bool = None,
    max_code_attempts: int = None,
    num_days_valid: int = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    debug_prn: bool = False,
):

    await self.toggle_mfa(
        is_enable_MFA=is_enable_MFA,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        suppress_update_self=True,
        debug_prn=debug_prn,
        
    )

    await self.set_max_code_attempts(
        max_code_attempts=max_code_attempts,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        suppress_update_self=True,
        debug_prn=debug_prn,
    )

    await self.set_num_days_valid(
        num_days_valid=num_days_valid,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        suppress_update_self=True,
        debug_prn=debug_prn,
    )

    return await self.get(debug_api = debug_api, return_raw = return_raw)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount_Default.ipynb.

# %% auto 0
__all__ = ['Account_CanIModify', 'UpsertAccount_MatchCriteria', 'DomoAccounConfig_MissingFields', 'DomoAccount_Default',
           'AccountClass_CRUD_Error']

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 3
from domolibrary.routes.account import (
    ShareAccount,
    Account_Share_Error,
    ShareAccount_V1_AccessLevel,
    ShareAccount_AccessLevel,
    )

from domolibrary.classes.DomoAccount_Config import (
    AccountConfig_UsesOauth,
    DomoAccount_Config,
    AccountConfig,
)

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 4
from dataclasses import dataclass, field
from typing import Any, List
import asyncio
import httpx
import datetime as dt


from nbdev.showdoc import patch_to

import domolibrary.utils.convert as cd
import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
from domolibrary.client.DomoEntity import DomoEntity
import domolibrary.routes.account as account_routes

import domolibrary.utils.chunk_execution as dmce

import domolibrary.classes.DomoAccess as dmas

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 7
class Account_CanIModify(dmde.ClassError):
    def __init__(self, account_id, domo_instance):
        super().__init__(
            message="`DomoAccount.is_admin_summary` must be `False` to proceed.  Either set the value explicity, or retrieve the account instance using `DomoAccount.get_by_id()`",
            domo_instance=domo_instance,
            entity_id=account_id,
        )


class UpsertAccount_MatchCriteria(dmde.ClassError):
    def __init__(self, domo_instance):
        super().__init__(
            message="must pass an account_id or account_name to UPSERT",
            domo_instance=domo_instance,
        )

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 9
class DomoAccounConfig_MissingFields(dmde.ClassError):
    def __init__(self, domo_instance, missing_keys, account_id):
        super().__init__(
            domo_instance=domo_instance,
            message=f"{account_id} config class definition is missing the following keys - {', '.join(missing_keys)} extend the AccountConfig",
        )


@dataclass
class DomoAccount_Default(DomoEntity):
    id: int
    auth: dmda.DomoAuth = field(repr=False)

    name: str = None
    data_provider_type: str = None

    created_dt: dt.datetime = None
    modified_dt: dt.datetime = None

    owners: List[Any] = None  # DomoUser or DomoGroup

    is_admin_summary: bool = True

    Config: DomoAccount_Config = field(repr=False, default=None)
    Access: dmas.DomoAccess_Account = field(repr=False, default=None)

    def __post_init__(self):
        self.id = int(self.id)

        self.Access = dmas.DomoAccess_Account._from_parent(
            parent=self,
        )

    def display_url(self):
        """returns the URL to the account in Domo"""
        return f"{self.auth.domo_instance}/datacenter/accounts"

    @classmethod
    def _default_from_json(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: dmda.DomoAuth = None,
        **kwargs,
    ):
        """converts data_v1_accounts API response into an accounts class object"""

        dd = util_dd.DictDot(obj)

        return cls(
            id=dd.id or dd.databaseId,
            name=dd.displayName,
            data_provider_type=dd.dataProviderId or dd.dataProviderType,
            created_dt=cd.convert_epoch_millisecond_to_datetime(
                dd.createdAt or dd.createDate
            ),
            modified_dt=cd.convert_epoch_millisecond_to_datetime(
                dd.modifiedAt or dd.lastModified
            ),
            auth=auth,
            is_admin_summary=is_admin_summary,
            owners=dd.owners,
            raw=obj,
            **kwargs,
        )

    @classmethod
    def _class_from_json(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: dmda.DomoAuth = None,
        **kwargs,
    ):
        return cls._default_from_json(
            obj=obj,
            is_admin_summary=is_admin_summary,
            auth=auth,
        )

    @classmethod
    def _from_json(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: dmda.DomoAuth = None,
        **kwargs,
    ):
        return cls._class_from_json(
            obj=obj, is_admin_summary=is_admin_summary, auth=auth, raw=obj, **kwargs
        )

    def _update_self(self, new_class, skip_props: list[str] = None):
        for key, value in new_class.__dict__.items():
            if key in skip_props:
                continue

            setattr(self, key, value)

        return True

    def _test_missing_keys(self, res_obj, config_obj):
        return [r_key for r_key in res_obj.keys() if r_key not in config_obj.keys()]

    async def _get_config(
        self,
        session=None,
        return_raw: bool = False,
        debug_api: bool = None,
        auth: dmda.DomoAuth = None,
        debug_num_stacks_to_drop=2,
        is_unmask: bool = True,
        is_suppress_no_config: bool = False,  # can be used to suppress cases where the config is not defined, either because the account_config is OAuth, and therefore not stored in Domo OR because the AccountConfig class doesn't cover the data_type
    ):
        if not self.data_provider_type:
            res = await account_routes.get_account_by_id(
                auth=self.auth,
                account_id=self.id,
                session=session,
                debug_api=debug_api,
                parent_class=self.__class__.__name__,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )

            self.data_provider_type = res.response["dataProviderType"]

        res = await account_routes.get_account_config(
            auth=auth or self.auth,
            account_id=self.id,
            session=session,
            debug_api=debug_api,
            is_unmask=is_unmask,
            data_provider_type=self.data_provider_type,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        config_fn = AccountConfig(self.data_provider_type).value

        try:
            self.Config = config_fn._from_json(
                obj=res.response, data_provider_type=self.data_provider_type
            )

        except dmde.DomoError as e:
            if not is_suppress_no_config:
                raise e from e
            print(e)

        if self.Config and self.Config.to_json() != {}:

            self._test_missing_keys(
                res_obj=res.response, config_obj=self.Config.to_json()
            )

        return self.Config

    @classmethod
    async def get_by_id(
        cls,
        auth: dmda.DomoAuth,
        account_id: int,
        is_suppress_no_config: bool = True,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        is_use_default_account_class=False,
        is_unmask=True,
        **kwargs,
    ):
        """retrieves account metadata and attempts to retrieve config"""

        res = await account_routes.get_account_by_id(
            auth=auth,
            account_id=account_id,
            session=session,
            debug_api=debug_api,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            is_unmask=is_unmask,
        )

        if return_raw:
            return res
        
        obj = res.response

        acc = cls._from_json(
            obj=obj,
            auth=auth,
            is_admin_summary=False,
            is_use_default_account_class=is_use_default_account_class,
            **kwargs,
        )

        await acc._get_config(
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            is_suppress_no_config=is_suppress_no_config,
        )

        return acc

    @classmethod
    async def get_entity_by_id(cls, entity_id, **kwargs):
        """Alias for get_by_id"""
        return await cls.get_by_id(account_id=entity_id, **kwargs)

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 15
@patch_to(DomoAccount_Default, cls_method=True)
async def create_account(
    cls: DomoAccount_Default,
    account_name: str,
    config: DomoAccount_Config,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2
):

    res = await account_routes.create_account(
        account_name=account_name,
        data_provider_type=config.data_provider_type,
        auth=auth,
        config_body=config.to_json(),
        debug_api=debug_api,
        session=session,
        parent_class = cls.__name__,
        debug_num_stacks_to_drop = debug_num_stacks_to_drop
    )

    if return_raw:
        return res

    acc = await cls.get_by_id(auth=auth, account_id=res.response.get("id"))
    acc.Config = config
    return acc


@patch_to(DomoAccount_Default)
async def update_name(
    self: DomoAccount_Default,
    account_name: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    auth = auth or self.auth

    res = await account_routes.update_account_name(
        auth=auth,
        account_id=self.id,
        account_name=account_name or self.name,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success and self.is_admin_summary:
        raise Account_CanIModify(account_id=self.id, domo_instance=auth.domo_instance)

    new_acc = await DomoAccount_Default.get_by_id(auth=auth, account_id=self.id)

    self._update_self(new_class = new_acc, skip_props = ['Config'])

    return self


@patch_to(DomoAccount_Default)
async def delete_account(
    self: DomoAccount_Default,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
    parent_class=None,
):
    auth = auth or self.auth

    res = await account_routes.delete_account(
        auth=auth,
        account_id=self.id,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success and self.is_admin_summary:
        raise Account_CanIModify(account_id=self.id, domo_instance=auth.domo_instance)

    return res

# %% ../../nbs/classes/50_DomoAccount_Default.ipynb 18
class AccountClass_CRUD_Error(dmde.ClassError):
    def __init__(self, cls_instance, message):
        super().__init__(cls_instance=cls_instance, message = message)


@patch_to(DomoAccount_Default)
async def update_config(
    self: DomoAccount_Default,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    config: DomoAccount_Config = None,
    is_suppress_no_config=False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    is_update_config: bool = False,  # if calling the Api, Domo will send encrypted values (astericks) in most cases it's best not to try to retrieve updated values from the API
):
    auth = auth or self.auth

    if config:
        self.Config = config

    if not self.Config:
        raise AccountClass_CRUD_Error(cls_instance= self, message = "unable to update account - no domo_account.Config not provided")

    res = await account_routes.update_account_config(
        auth=auth,
        account_id=self.id,
        config_body=self.Config.to_json(),
        debug_api=debug_api,
        session=session,
    )

    # await asyncio.sleep(3)

    # new_acc = await DomoAccount_Default.get_by_id(auth=auth, account_id=self.id)

    # self._update_self(new_class = new_acc, skip_props = ['Config'])

    if return_raw:
        return res

    if not res.is_success and self.is_admin_summary:
        raise Account_CanIModify(account_id=self.id, domo_instance=auth.domo_instance)

    if not is_update_config:
        return self.Config

    await asyncio.sleep(3)

    await self._get_config(
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        is_suppress_no_config=is_suppress_no_config,
    )

    return self.Config



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/workflows.ipynb.

# %% auto 0
__all__ = ['get_workflow', 'generate_trigger_workflow_body', 'trigger_workflow', 'get_workflow_trigger_history',
           'get_workflow_executions']

# %% ../../nbs/routes/workflows.ipynb 1
import domolibrary.client.get_data as gd
import domolibrary.client.DomoError as dmde
import domolibrary.client.DomoAuth as dmda
import httpx

# %% ../../nbs/routes/workflows.ipynb 5
@gd.route_function
async def get_workflow(
    auth: dmda.DomoAuth,
    model_id: str,
    version_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):

    url = f"https://{auth.domo_instance}.domo.com/api/workflow/v1/models/{model_id}/versions/{version_id}"
    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res)

    return res

# %% ../../nbs/routes/workflows.ipynb 8
def generate_trigger_workflow_body(
    starting_tile, model_id, version_id, execution_params: dict
):
    return {
        "messageName": starting_tile,
        "version": version_id,
        "modelId": model_id,
        "data": execution_params,
    }


@gd.route_function
async def trigger_workflow(
    auth: dmda.DomoAuth,
    starting_tile: str,
    model_id: str,
    version_id: str,
    execution_parameters: dict = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    body = generate_trigger_workflow_body(
        starting_tile=starting_tile,
        model_id=model_id,
        execution_params=execution_parameters,
        version_id=version_id,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/workflow/v1/instances/message"

    res = await gd.get_data(
        method="POST",
        url=url,
        body=body,
        auth=auth,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res)

    return res

# %% ../../nbs/routes/workflows.ipynb 12
@gd.route_function
async def get_workflow_trigger_history(
    auth: dmda.DomoAuth,
    model_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):

    url = f"https://{auth.domo_instance}.domo.com/api/workflow/v2/executions/{model_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res)

    return res

# %% ../../nbs/routes/workflows.ipynb 14
@gd.route_function
async def get_workflow_executions(
    auth: dmda.DomoAuth,
    model_id: str,
    version_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False
):
    params = {
        "modelId": model_id,
        #   "triggerTypes" : "ALERT,API,APP_STUDIO,CUSTOM_APP,MANUAL,TIMER,WORKFLOW"
        "version": version_id,
        # "status"  : "IN_PROGRESS"
    }

    url = f"https://{auth.domo_instance}.domo.com/api/workflow/v2/executions"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        debug_api=debug_api,
        params=params,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res)
    
    if return_raw:
        return res

    res.response = res.response.get("executions", [])

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoMembership.ipynb.

# %% ../../nbs/classes/50_DomoMembership.ipynb 2
from __future__ import annotations
from typing import List, Any
from dataclasses import dataclass, field

import httpx

from abc import abstractmethod

from nbdev.showdoc import patch_to

import domolibrary.utils.chunk_execution as dmce

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
from domolibrary.client.DomoEntity import DomoSubEntity, Entity_Relation

import domolibrary.routes.group as group_routes


# %% auto 0
__all__ = ['UpdateMembership', 'Membership_Entity', 'Membership', 'GroupMembership']

# %% ../../nbs/classes/50_DomoMembership.ipynb 5
class UpdateMembership(dmde.ClassError):
    def __init__(self, cls_instance, member_name=None, entity_id=None):
        super().__init__(
            entity_id=entity_id,
            cls_instance=cls_instance,
            message=f"unable to alter mmbership {member_name if member_name else ''}",
        )


@dataclass
class Membership_Entity(Entity_Relation):

    def to_json(self):
        import domolibrary.classes.DomoUser as dmdu
        import domolibrary.classes.DomoGroup as dmdg

        if isinstance(self.entity, dmdu.DomoUser):
            return {"type": "USER", "id": str(self.entity.id)}

        if isinstance(self.entity, dmdg.DomoGroup):
            return {"type": "GROUP", "id": int(self.entity.id)}

        raise UpdateMembership(cls_instance=self, entity_id=self.entity.id)


@dataclass
class Membership(DomoSubEntity):

    owners: List[str] = field(default_factory=lambda: [])
    members: List[str] = field(default_factory=lambda: [])

    _add_member_ls: List[str] = field(default_factory=lambda: [])
    _remove_member_ls: List[str] = field(default_factory=lambda: [])

    _add_owner_ls: List[str] = field(default_factory=lambda: [])
    _remove_owner_ls: List[str] = field(default_factory=lambda: [])


    @abstractmethod
    async def get_owners(self):
        pass

    @abstractmethod
    async def get_members(self):
        pass

    @abstractmethod
    async def update(self):
        pass

    def _add_to_list(self, member, list_to_update, relation_type):

        if not isinstance(
            member,
            Membership_Entity,
        ):

            member = Membership_Entity(entity=member, relation_type=relation_type, auth = member.auth)

        if member not in list_to_update:
            list_to_update.append(member)

    def _add_member(self, member):
        return self._add_to_list(member, self._add_member_ls, relation_type = "member")

    def _remove_member(self, member, is_keep_system_group=True):
        """does not remove system groups"""
        import domolibrary.classes.DomoGroup as dmg

        if is_keep_system_group and isinstance(member, dmg.DomoGroup) and member.type == "system":
            return

        return self._add_to_list(member, self._remove_member_ls, relation_type = "member")

    def _add_owner(self, member):
        return self._add_to_list(member, self._add_owner_ls, relation_type = "owner")

    def _remove_owner(self, member, is_keep_system_group=True):
        """does not remove system groups"""
        import domolibrary.classes.DomoGroup as dmg

        if is_keep_system_group and isinstance(member, dmg.DomoGroup) and member.type == "system":
            return  

        return self._add_to_list(member, self._remove_owner_ls, relation_type = "owner")

    def _reset_obj(self):
        self._add_member_ls = []
        self._remove_member_ls = []

        self._add_owner_ls = []
        self._remove_owner_ls = []

    async def _extract_domo_groups_from_list(
        self, entity_ls, session: httpx.AsyncClient
    ):
        import domolibrary.classes.DomoGroup as dmg

        return await dmce.gather_with_concurrency(
            *[
                dmg.DomoGroup.get_by_id(
                    group_id=obj.get("groupId") or obj.get("id"),
                    auth=self.auth,
                    session=session,
                )
                for obj in entity_ls
                if obj.get("type") == "GROUP" or obj.get("groupId")
            ],
            n=60,
        )

    async def _extract_domo_users_from_list(
        self, entity_ls, session: httpx.AsyncClient = None
    ):
        import domolibrary.classes.DomoUser as dmu

        return await dmce.gather_with_concurrency(
            *[
                dmu.DomoUser.get_by_id(
                    user_id=obj.get("userId") or obj.get("id"),
                    session=session,
                    auth=self.auth,
                )
                for obj in entity_ls
                if obj.get("type") == "USER" or obj.get("userId")
            ],
            n=10,
        )

    async def _extract_domo_entities_from_list(
        self,
        entity_ls, 
        relation_type,
        session: httpx.AsyncClient = None
    ):
        session = session or httpx.AsyncClient()

        domo_groups = await self._extract_domo_groups_from_list(
            entity_ls, session=session
        )
        domo_users = await self._extract_domo_users_from_list(
            entity_ls, session=session
        )

        return [Membership_Entity(entity=entity, relation_type=relation_type, auth = entity.auth) for entity in domo_groups + domo_users]

    @staticmethod
    def _list_to_json(entity_ls):
        return [parent.to_json() for parent in entity_ls]

# %% ../../nbs/classes/50_DomoMembership.ipynb 7
@dataclass
class GroupMembership(Membership):

    async def get_owners(
        self,
        return_raw: bool = False,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):

        res = await group_routes.get_group_owners(
            group_id=self.parent_id,
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
        if return_raw:
            return res

        owners = await self._extract_domo_entities_from_list(res.response, relation_type = 'OWNER', session = session)
        self.owners = owners

        if self.parent:
            self.parent.owner_id_ls = [owner.entity.id for owner in self.owners]
            self.parent.owner_ls = [owner.entity for owner in self.owners]

        return self.owners

    async def update(
        self,
        update_payload : dict = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        res = await group_routes.update_group_membership(
            auth=self.auth,
            update_payload=update_payload,
            group_id=self.parent_id,
            add_member_arr=self._list_to_json(self._add_member_ls),
            remove_member_arr=self._list_to_json(self._remove_member_ls),
            add_owner_arr=self._list_to_json(self._add_owner_ls),
            remove_owner_arr=self._list_to_json(self._remove_owner_ls),
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        self._reset_obj()

        return res

    async def get_members(
        self,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
    ):

        res = await group_routes.get_group_membership(
            group_id=self.parent_id,
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        members = await self._extract_domo_entities_from_list(res.response, relation_type = 'MEMBER', session = session)
        self.members = members

        if self.parent:
            self.parent.members_id_ls = [member.entity.id for member in self.members]
            self.parent.members_ls = [member.entity for member in self.members]

        return self.members

# %% ../../nbs/classes/50_DomoMembership.ipynb 14
@patch_to(GroupMembership)
async def add_members(
    self: GroupMembership,
    add_user_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    self._reset_obj()

    for domo_user in add_user_ls:    
        self._add_member(domo_user)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if return_raw:
        return res

    return await self.get_members(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )


@patch_to(GroupMembership)
async def remove_members(
    self: GroupMembership,
    remove_user_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    self._reset_obj()

    for domo_user in remove_user_ls:
        self._remove_member(domo_user)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if return_raw:
        return res

    return await self.get_members(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )


@patch_to(GroupMembership)
async def set_members(
    self: GroupMembership,
    user_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    self._reset_obj()

    user_ls = [
        Membership_Entity(entity=user, relation_type = 'MEMBER',
                          auth = self.auth) for user in user_ls
    ]

    memberships = await self.get_members(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    for domo_user in user_ls:
        self._add_member(domo_user)

    for me in memberships:
        if me not in user_ls:
            self._remove_member(me)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )
    if return_raw:
        return res

    return await self.get_members(
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        session=session,
    )

# %% ../../nbs/classes/50_DomoMembership.ipynb 16
@patch_to(GroupMembership)
async def add_owners(
    self: GroupMembership,
    add_owner_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    self._reset_obj()

    for domo_user in add_owner_ls:
        self._add_owner(domo_user)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    return await self.get_owners(
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )


@patch_to(GroupMembership)
async def remove_owners(
    self: GroupMembership,
    remove_owner_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    self._reset_obj()

    for domo_user in remove_owner_ls:
        self._remove_owner(domo_user)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if return_raw:
        return res

    return await self.get_owners(
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )


@patch_to(GroupMembership)
async def set_owners(
    self: GroupMembership,
    owner_ls: List[Any],
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    import domolibrary.classes.DomoGroup as dmdg

    self._reset_obj()
    
    owner_ls = [
        Membership_Entity(entity=owner,
                          relation_type='OWNER',
                          auth=self.auth) for owner in owner_ls
    ]
    
    membership = await self.get_owners()

    for domo_entity in owner_ls:
        self._add_owner(domo_entity)

    for oe in membership:

        # open accounts must have themselves as an owner
        if (
            self.parent
            and self.parent.type == "open"
            and self.parent.id == oe.entity.id
            and isinstance(oe.entity, dmdg.DomoGroup)
        ):
            self._add_owner(oe)
            continue

        if isinstance(oe.entity, dmdg.DomoGroup) and oe.entity.is_system:
            self._add_owner(oe)
            continue

        if oe.entity not in owner_ls:
            self._remove_owner(oe)

    res = await self.update(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )

    if return_raw:
        return res

    return await self.get_owners(
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
    )


@patch_to(GroupMembership)
async def add_owner_manage_all_groups_role(
    self: GroupMembership, debug_api: bool = False, session: httpx.AsyncClient = None
):
    import domolibrary.classes.DomoGroup as dmg

    domo_groups = dmg.DomoGroups(auth=self.auth)

    grant_group = await domo_groups.search_by_name(
        group_name="Grant: Manage all groups",
        is_hide_system_groups=False,
        session=session,
        debug_api=debug_api,
    )

    await self.add_owners(
        add_owner_ls=[grant_group], session=session, debug_api=debug_api
    )

    return await self.get_owners(session = session)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccess.ipynb.

# %% auto 0
__all__ = ['Access_Config_Error', 'Access_Entity', 'DomoAccess', 'DomoAccess_Account', 'DomoAccess_OAuth']

# %% ../../nbs/classes/50_DomoAccess.ipynb 3
from domolibrary.routes.account import (
    ShareAccount_AccessLevel,
    ShareAccount_V1_AccessLevel,
    ShareAccount,
    Account_Share_Error,
)

# %% ../../nbs/classes/50_DomoAccess.ipynb 4
from dataclasses import dataclass, field
from typing import Any, List

from abc import abstractmethod

import httpx

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
import domolibrary.routes.account as account_routes
from domolibrary.client.DomoEntity import DomoSubEntity, Entity_Relation

import domolibrary.utils.chunk_execution as dmce

# %% ../../nbs/classes/50_DomoAccess.ipynb 7
class Access_Config_Error(dmde.ClassError):
    def __init__(self, cls_instance, account_id, message):
        super().__init__(
            cls_instance=cls_instance, message=message, entity_id=account_id
        )

@dataclass
class Access_Entity(Entity_Relation):
   
   relation_type : ShareAccount = None
   
   
        
@dataclass
class DomoAccess(DomoSubEntity):
    auth: dmda.DomoAuth = field(repr=False)

    share_enum: ShareAccount = field(repr = False)

    accesslist: List[Access_Entity] = field(default_factory=lambda: []) # can include users or groups

    accesslist_all_users: List[Any] = field(default_factory=lambda: [])  # DomoUser

    def __post_init__(self):
        super().__post_init__()

        if not issubclass(self.share_enum, ShareAccount):
            
            raise Access_Config_Error(
                cls_instance=self,
                account_id=self.parent_id,
                message="Share enum must be a subclass of ShareAccount.",
            )

    
    async def get(self, *args, **kwargs):
        raise NotImplementedError(
            f"{self.__class__.__name__}.get() must be implemented in a subclass."
        )


    async def get_all_users(
        self,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 1,
        session: httpx.AsyncClient = None,
    ):

        import domolibrary.classes.DomoUser as dmdu
        import domolibrary.classes.DomoGroup as dmdg

        await self.get(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        for access in self.accesslist:
            entity = access.entity

            if isinstance(entity, dmdu.DomoUser):
                if entity not in self.accesslist_all_users:
                    self.accesslist_all_users.append(entity)

            if isinstance(entity, dmdg.DomoGroup):
                domo_users = await access.entity.Membership.get_members()

                [
                    self.accesslist_all_users.append(user)
                    for user in domo_users
                    if user not in self.accesslist_all_users
                ]

        return self.accesslist_all_users

    async def share(
        self,
        user_id=None,
        group_id=None,
        domo_group=None,
        domo_user=None,
        relation_type: ShareAccount = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        relation_type = relation_type or self.share_enum.default

        user_id = user_id or domo_user and domo_user.id
        group_id = group_id or domo_group and domo_group.id

        if not user_id and not group_id:
            raise Access_Config_Error(
                cls_instance=self,
                account_id=self.parent_id,
                message="Either user_id or group_id must be provided.",
            )

        share_payload = relation_type.generate_payload(
            user_id=user_id,
            group_id=group_id
        )

        res = await relation_type.share(
            auth=self.auth,
            account_id=self.parent_id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        return res
    
    async def upsert_share(
        self,
        user_id=None,
        group_id=None,
        domo_group=None,
        domo_user=None,
        relation_type: ShareAccount = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        user_id=user_id or (domo_user and domo_user.id)
        group_id=group_id or (domo_group and domo_group.id)

        await self.get(session = session, debug_api = debug_api, debug_num_stacks_to_drop = debug_num_stacks_to_drop + 1, )
        
        is_already_exists = all([( member.entity.id == user_id or member.entity.id == group_id) and relation_type == member.relation_type  for member in self.accesslist])
            

        if is_already_exists:
            return False

        return await self.share(
            user_id = user_id,
            group_id = group_id,
            relation_type= relation_type,
            debug_api= debug_api,
            debug_num_stacks_to_drop= debug_num_stacks_to_drop,
            session = session
        )    
        

# %% ../../nbs/classes/50_DomoAccess.ipynb 9
@dataclass
class DomoAccess_Account(DomoAccess):
    version: int = None  # api version - aligns to feature switch

    share_enum: ShareAccount = field(repr = False, default = ShareAccount_AccessLevel)

    def __post_init__(self):

        super().__post_init__()
        
        if isinstance(self.share_enum , ShareAccount_AccessLevel):
            self.version = 2
        elif isinstance(self.share_enum , ShareAccount_V1_AccessLevel):
            self.version = 1
        
        return True

    async def get(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        """assumes v2 access api"""

        res = await account_routes.get_account_accesslist(
            auth=self.auth,
            account_id=self.parent_id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.accesslist = await dmce.gather_with_concurrency(
            *[
                Access_Entity.from_user_id(
                    user_id=obj["id"],
                    auth=self.auth,
                    relation_type=self.share_enum.get(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "USER"
            ],
            *[
                Access_Entity.from_group_id(
                    group_id=obj["id"],
                    auth=self.auth,
                    relation_type=self.share_enum.get(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "GROUP"
            ],
            n=10,
        )
        return self.accesslist

# %% ../../nbs/classes/50_DomoAccess.ipynb 17
@dataclass
class DomoAccess_OAuth(DomoAccess):
    share_enum: ShareAccount = field(
        repr=False, 
        default=ShareAccount_AccessLevel
    )

    async def get(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        """assumes v2 access api"""

        res = await account_routes.get_oauth_account_accesslist(
            auth=self.auth,
            account_id=self.parent_id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.accesslist = await dmce.gather_with_concurrency(
            *[
                Access_Entity.from_user_id(
                    user_id=obj["id"],
                    auth=self.auth,
                    relation_type=self.share_enum.get(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "USER"
            ],
            *[
                Access_Entity.from_group_id(
                    group_id=obj["id"],
                    auth=self.auth,
                    relation_type=self.share_enum.get(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "GROUP"
            ],
            n=10,
        )
        return self.accesslist



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/filesets.ipynb.

# %% auto 0
__all__ = ['EmbedData_Type', 'create_filesets_index', 'embed_image', 'get_fileset_by_id', 'search_fileset_files',
           'get_data_file_by_id']

# %% ../../nbs/routes/filesets.ipynb 1
import domolibrary.client.get_data as gd
import domolibrary.client.DomoError as dmde
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.ResponseGetData as rgd
from typing import Optional, Literal, get_args

import httpx

# %% ../../nbs/routes/filesets.ipynb 5
@gd.route_function
async def create_filesets_index(
    auth: dmda.DomoAuth,
    index_id,
    embedding_model: str = "domo.domo_ai.domo-embed-text-multilingual-v1:cohere",
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
):
    """Creates a new vectorDB index."""

    url = f"{auth.domo_instance}.domo.com/api/recall/v1/indexes"

    payload = {
        "indexId": index_id,
        "embeddingModel": embedding_model,
    }
    res = await gd.get_data(
        url,
        method="POST",
        body=payload,
        auth=auth,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res)

    return res

# %% ../../nbs/routes/filesets.ipynb 6
EmbedData_Type = Literal["base64"]
    
@gd.route_function
async def embed_image(
    auth: dmda.DomoAuth,
    body : dict = None,
    image_data: str = None,
    media_type: str=None,
    data_type : EmbedData_Type = "base64",
    model: str = "domo.domo_ai",
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: Optional[httpx.AsyncClient] = None,
) -> rgd.ResponseGetData:
    """
    Create an embedding for a base64 encoded image using Domo's AI services.
    """
    # Utility function is_valid_base64_image should be called by the orchestrator before this.
    # This route function assumes valid base64 data.

    api_url = f"https://{auth.domo_instance}.domo.com/api/ai/v1/embedding/image"
    
    body = body or{
        "input": [
            {"type": "", "mediaType": "", "data": ""}
        ],
        "model": "",
    }
    if image_data:
        body['input'][0].update({"data": image_data})

    if media_type:
        body['input'][0].update({"mediaType": media_type})
    
    if data_type:
        body['input'][0].update({"type": data_type})
    
    if model:
        body.update({'model': model})

    res = await gd.get_data(
        auth=auth,
        url=api_url,
        method="POST",
        body=body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=parent_class,
    )
    if not res.is_success:
        raise dmde.RouteError(res=res)
    
    return res


# %% ../../nbs/routes/filesets.ipynb 10
@gd.route_function
async def get_fileset_by_id(
    auth: dmda.DomoAuth,
    fileset_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/files/v1/filesets/{fileset_id}"
    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise dmde.RouteError(res=res, debug_api=debug_api)

    return res

# %% ../../nbs/routes/filesets.ipynb 13
@gd.route_function
async def search_fileset_files(
    auth: dmda.DomoAuth,
    domo_fileset_id: str,
    body: dict = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/files/v1/filesets/{domo_fileset_id}/files/search?directoryPath=&immediateChildren=true"

    if not body:
        # default body will pull all files within the given fileset_id
        body = {
            "fieldSort": [{"field": "created", "order": "DESC"}],
            "filters": [],
            "dateFilters": [],
        }

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )
    if not res.is_success:
        raise dmde.RouteError(res=res)

    return res

# %% ../../nbs/routes/filesets.ipynb 15
@gd.route_function
async def get_data_file_by_id(
    auth: dmda.DomoAuth,
    file_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: Optional[httpx.AsyncClient] = None,
) -> rgd.ResponseGetData:
    """
    Retrieves the content of a data file from Domo.
    """
    url = f"https://{auth.domo_instance}.domo.com/data/v1/data-files/{file_id}"
    res = gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=parent_class,
    )
    if not res.is_success:
        raise dmde.RouteError(res=res)
    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/codeengine_package_analyzer.ipynb.

# %% auto 0
__all__ = ['deploy_package', 'CodeEngine_InvalidPackage', 'create_package', 'CodeEngine_PackageNotFound',
           'CodeEngine_PackageAlreadyDeployed', 'update_package', 'upsert_package', 'remove_javascript_comments',
           'parse_javascript_between_functions', 'parse_javascript_between_module', 'remove_after_closing_parens',
           'parse_javascript', 'parse_python', 'parse_functions_factory', 'extract_functions']

# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 2
import httpx

from domolibrary.client.DomoEntity import DomoEnum
from typing import List, Callable
from functools import partial

import re

import domolibrary.client.get_data as gd
import domolibrary.client.DomoError as dmde
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
# from domolibrary.utils.CodeEngineUtils import (
#     CodeEngineScriptAnalyzer,
#     CodeEnginePackageBuilder
# )
from typing import Dict, Any

# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 5
@gd.route_function
async def deploy_package(
    package_id,
    version,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    
    if not package_id or not version:
        raise CodeEngine_FunctionCallError(
            message= f"Package ID {package_id or 'not provided '} and version {version or ' not provided  '} must be provided.",
            auth=auth,
        )
    
    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages/{package_id}/versions/{version}/release"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CodeEngine_API_Error(res=res)
    
    return res

# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 7
class CodeEngine_InvalidPackage(dmde.DomoError):
    def __init__(self, message: str, auth: dmda.DomoAuth):
        super().__init__(message=message, domo_instance=auth.domo_instance)

@gd.route_function
async def create_package(
    package: Dict[str, Any] | CodeEnginePackageBuilder,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:

    if isinstance(package, dict):
        package = CodeEnginePackageBuilder().from_dict(data=package)
    try:
        package = package.build()

    except Exception as e:
        raise CodeEngine_InvalidPackage(
            message=f"Failed to build package: {str(e)}",
            auth=auth,
        )

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        body=package
    )

    if not res.is_success:
        raise CodeEngine_API_Error(res=res)
    
    return res

# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 10
class CodeEngine_PackageNotFound(dmde.DomoError):
    def __init__(self, message: str, auth: dmda.DomoAuth):
        super().__init__(message=message, domo_instance=auth.domo_instance)

class CodeEngine_PackageAlreadyDeployed(dmde.DomoError):
    def __init__(self, message: str, auth: dmda.DomoAuth):
        super().__init__(message=message, domo_instance=auth.domo_instance)

@gd.route_function
async def update_package(
    package: Dict[str, Any] | CodeEnginePackageBuilder,
    auth: dmda.DomoAuth,
    check_different: bool = True,
    create_new_version: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:

    if isinstance(package, dict):
        package = CodeEnginePackageBuilder().from_dict(data=package)
    try:
        package = package.build()
    except Exception as e:
        raise CodeEngine_InvalidPackage(
            message=f"Failed to build package: {str(e)}",
            auth=auth,
        )

    pkg_id = package['id']
    pkg_ver = package['version']

    # Ensure package exists (let not-found propagate)
    if debug_api:
        print(f"Checking for package with ID {pkg_id}...")
    await get_codeengine_package_by_id(
        pkg_id,
        auth=auth,
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not create_new_version:
        if debug_api:
            print(f"Retrieving full package information for ID {pkg_id} and version {pkg_ver}...")
        try:
            codeengine_full_package = await get_codeengine_package_by_id_and_version(
                pkg_id,
                pkg_ver,
                auth=auth,
                params={"parts": "code"},
                debug_api=debug_api,
                session=session,
                parent_class=parent_class,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )

            if codeengine_full_package.response.get('released') is not None:
                raise CodeEngine_PackageAlreadyDeployed(
                    message=f"Package with ID {pkg_id} and version {pkg_ver} is already deployed",
                    auth=auth,
                )

            if check_different:
                existing = codeengine_full_package.response or {}
                try:
                    new_code = package.get('code') if isinstance(package, dict) else getattr(package, 'code', None)
                except Exception:
                    new_code = None

                if existing.get('code') == new_code:
                    if debug_api:
                        print(f"Package with ID {pkg_id} and version {pkg_ver} is identical to existing; skipping update.")
                    return codeengine_full_package

        except CodeEngine_PackageNotFound:
            # Version not found -> proceed to update (POST)
            pass

        if debug_api:
            print(f"Updating package with ID {pkg_id} and version {pkg_ver}...")

    url = f"https://{auth.domo_instance}.domo.com/api/codeengine/v2/packages"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        body=package,
    )

    if not res.is_success:
        raise CodeEngine_API_Error(res=res)

    return res

# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 13
@gd.route_function
async def upsert_package(
    package: Dict[str, Any] | CodeEnginePackageBuilder,
    auth: dmda.DomoAuth,
    check_different: bool = True,
    create_new_version: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:

    if isinstance(package, dict):
        pkg_builder = CodeEnginePackageBuilder().from_dict(data=package)
    else:
        pkg_builder = package

    try:
        pkg = pkg_builder.build()
    except Exception as e:
        raise CodeEngine_InvalidPackage(
            message=f"Failed to build package: {str(e)}",
            auth=auth,
        )

    pkg_id = pkg['id']
    result = None
    if not pkg_id:
        if debug_api:
            print('No Package ID found, creating new package...')
        result = await create_package(
            pkg,
            auth=auth,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
        return result

    try:
        await get_codeengine_package_by_id(
            pkg_id,
            auth=auth,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
    except CodeEngine_PackageNotFound:
        result = await create_package(
            pkg,
            auth=auth,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
        return result

    result = await update_package(
        pkg_builder,
        auth=auth,
        check_different=check_different,
        create_new_version=create_new_version,
        debug_api=debug_api,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )
    return result


# %% ../../nbs/routes/codeengine_package_analyzer.ipynb 17
def remove_javascript_comments(code, **kwargs):
    pattern = r"(/\*\*(.*?)\*/)"
    return re.sub(pattern, "", code, flags=re.DOTALL)


def parse_javascript_between_functions(code, function_name):
    # pattern = r"((async\s)?function\s+" + function_name + r"\(.*?\).*?)(async\s)?function\s"
    pattern = r"((async\s)?function\s+" + function_name + r"\(.*?\).*?}\n\n\n)"
    match = re.search(pattern, code, re.DOTALL)
    return match.group(0) if match else None


def parse_javascript_between_module(code, function_name):
    pattern = (
        r"((async\s)?function\s+" + function_name + r"\(.*?\).*?)(?=module.export)"
    )
    match = re.search(pattern, code, re.DOTALL)
    return match.group(0) if match else None


def remove_after_closing_parens(code, **kwargs):
    pattern = r"(.*\})"
    match = re.search(pattern, code, re.DOTALL)
    return match.group(0) if match else None


def parse_javascript(function_name, code):
    # remove comments
    code = remove_javascript_comments(code)

    # find everything until the next function definition
    match_str = parse_javascript_between_functions(
        code=code, function_name=function_name
    )

    if not match_str:
        match_str = parse_javascript_between_module(
            code=code, function_name=function_name
        )

    # remove everything after the last closing parens
    return remove_javascript_comments(match_str)


def parse_python(function_name, code):
    pattern = (
        r"(async\s+)?def\s+" + re.escape(function_name) + r"\s*\([^)]*\)\s*:\s*[^}]*"
    )
    match = re.search(pattern, code, re.DOTALL)
    return match.group(0)


class parse_functions_factory(DomoEnum):
    PYTHON = partial(parse_python)
    JAVASCRIPT = partial(parse_javascript)


def extract_functions(
    function_ls: List[dict],  # from API, function metadata
    code: str,  # from API, code string
    language: str,  # PYTHON or JAVASCRIPT, must align with parse_functions_factory enum
    parse_fn: Callable = None,  # function for exctracting functions must receive `code:str` and function_name:str`
):
    """helper function that handles parsing"""

    parse_fn = parse_fn or parse_functions_factory[language]

    res = []

    # return [{ **function, "code" : parse_fn.value(function_name=function["name"], code=code)}
    #     for function in function_ls
    # ]
    for function in function_ls:
        function_name = function["name"]
        # print(function_name)
        code_obj = {"code": parse_fn.value(function_name=function_name, code=code)}
        # print(code_obj)
        res.append({**function, **code_obj})

    return res
